from .time_window_controllers import StartAndDurationController, RangeController
from .group_and_sort_controllers import GroupAndSortController
from .misc import ProgressBar, make_trial_event_controller



import numpy as np


def dicts_exact_equal(first, second):
    """Return whether two dicts of arrays are exactly equal"""
    if first.keys() != second.keys():
        return False
    return all(np.array_equal(first[key], second[key]) for key in first)


def dicts_approximate_equal(first, second):
    """Return whether two dicts of arrays are roughly equal"""
    if first.keys() != second.keys():
        return False
    return all(np.allclose(first[key], second[key]) for key in first)



from typing import Iterable

import numpy as np

import ipywidgets as widgets

from hdmf.common import DynamicTable
from pynwb.misc import Units

from .base import lazy_tabs, render_dataframe, TimeIntervalsSelector
from .controllers import GroupAndSortController
from .misc import RasterWidget, PSTHWidget, RasterGridWidget, TuningCurveWidget
from .utils.pynwb import robust_unique
from .view import default_neurodata_vis_spec


class AllenRasterWidget(RasterWidget):
    def make_group_and_sort(self, group_by=None, control_order=False):
        return AllenRasterGroupAndSortController(
            self.units, group_by=group_by, control_order=control_order
        )


class AllenPSTHWidget(TimeIntervalsSelector):
    InnerWidget = PSTHWidget


class AllenRasterGridWidget(TimeIntervalsSelector):
    InnerWidget = RasterGridWidget


class AllenTuningCurveWidget(TimeIntervalsSelector):
    InnerWidget = TuningCurveWidget


class AllenRasterGroupAndSortController(GroupAndSortController):
    def get_groups(self):

        self.electrodes = self.dynamic_table.get_ancestor("NWBFile").electrodes

        groups = super().get_groups()
        for name in self.electrodes.colnames:
            if not name == "group":
                groups.update(**{name: np.unique(self.electrodes[name][:])})
        return groups

    def get_orderable_cols(self):
        units_orderable_cols = super().get_orderable_cols()
        candidate_cols = [
            x
            for x in self.electrodes.colnames
            if not (
                isinstance(self.electrodes[x][0], (Iterable, str))
            )
        ]
        return units_orderable_cols + [
            x for x in candidate_cols if len(robust_unique(self.electrodes[x][:])) > 1
        ]

    def get_group_vals(self, by, rows_select=()):
        if by is None:
            return None
        elif by in self.dynamic_table:
            return self.dynamic_table[by][:][rows_select]
        else:
            if self.electrodes is not None and by in self.electrodes:
                ids = self.electrodes.id[:]
                inds = [
                    np.argmax(ids == val)
                    for val in self.dynamic_table["peak_channel_id"][:]
                ]
                return self.electrodes[by][:][inds][rows_select]


def allen_show_dynamic_table(node: DynamicTable, **kwargs) -> widgets.Widget:
    if node.name == "electrodes":
        return allen_show_electrodes(node)
    return render_dataframe(node)


def allen_show_electrodes(node: DynamicTable):
    from ccfwidget import CCFWidget

    return lazy_tabs(dict(table=render_dataframe, CCF=CCFWidget), node)


def load_allen_widgets():
    default_neurodata_vis_spec[Units]["Session Raster"] = AllenRasterWidget
    default_neurodata_vis_spec[Units]["Grouped PSTH"] = AllenPSTHWidget
    default_neurodata_vis_spec[Units]["Raster Grid"] = AllenRasterGridWidget
    default_neurodata_vis_spec[Units]["Tuning Curves"] = AllenTuningCurveWidget
    # default_neurodata_vis_spec[DynamicTable] = allen_show_dynamic_table



import matplotlib.pyplot as plt

import plotly.graph_objects as go
from ipywidgets import widgets, fixed, Layout

from pynwb import TimeSeries
from pynwb.image import GrayscaleImage, ImageSeries, RGBImage

from tifffile import imread, TiffFile

from .base import fig2widget
from .controllers import StartAndDurationController
from .utils.timeseries import (
    get_timeseries_maxt,
    get_timeseries_mint,
    timeseries_time_to_ind,
)


class ImageSeriesWidget(widgets.VBox):
    """Widget showing ImageSeries."""

    def __init__(
        self,
        imageseries: ImageSeries,
        foreign_time_window_controller: StartAndDurationController = None,
        **kwargs
    ):
        super().__init__()
        self.imageseries = imageseries
        self.controls = {}
        self.out_fig = None

        # Set controller
        if foreign_time_window_controller is None:
            tmin = get_timeseries_mint(imageseries)
            if imageseries.external_file and imageseries.rate:
                tif = TiffFile(imageseries.external_file[0])
                tmax = imageseries.starting_time + len(tif.pages) / imageseries.rate
            else:
                tmax = get_timeseries_maxt(imageseries)
            self.time_window_controller = StartAndDurationController(tmax, tmin)
        else:
            self.time_window_controller = foreign_time_window_controller
        self.set_controls(**kwargs)

        # Make widget figure
        self.set_out_fig()

        self.children = [self.out_fig, self.time_window_controller]

    def time_to_index(self, time):
        if self.imageseries.external_file and self.imageseries.rate:
            return int((time - self.imageseries.starting_time) * self.imageseries.rate)
        else:
            return timeseries_time_to_ind(self.imageseries, time)

    def set_controls(self, **kwargs):
        self.controls.update(
            timeseries=fixed(self.imageseries), time_window=self.time_window_controller
        )
        self.controls.update({key: widgets.fixed(val) for key, val in kwargs.items()})

    def get_frame(self, idx):
        if self.imageseries.external_file is not None:
            return imread(self.imageseries.external_file, key=idx)
        else:
            return self.image_series.data[idx].T

    def set_out_fig(self):

        self.out_fig = go.FigureWidget(
            data=go.Heatmap(
                z=self.get_frame(0),
                colorscale="gray",
                showscale=False,
            )
        )
        self.out_fig.update_layout(
            xaxis=go.layout.XAxis(showticklabels=False, ticks=""),
            yaxis=go.layout.YAxis(
                showticklabels=False, ticks="", scaleanchor="x", scaleratio=1
            ),
        )

        def on_change(change):
            # Read frame
            frame_number = self.time_to_index(change["new"][0])
            image = self.get_frame(frame_number)
            self.out_fig.data[0].z = image

        self.controls["time_window"].observe(on_change)


def show_image_series(image_series: ImageSeries, neurodata_vis_spec: dict):
    if len(image_series.data.shape) == 3:
        return show_grayscale_image_series(image_series, neurodata_vis_spec)

    def show_image(index=0, mode="rgb"):
        fig, ax = plt.subplots(subplot_kw={"xticks": [], "yticks": []})
        image = image_series.data[index]
        if mode == "bgr":
            image = image[:, :, ::-1]
        ax.imshow(image.transpose([1, 0, 2]), cmap="gray", aspect="auto")
        fig.show()
        return fig2widget(fig)

    slider = widgets.IntSlider(
        value=0,
        min=0,
        max=image_series.data.shape[0] - 1,
        orientation="horizontal",
        continuous_update=False,
        description="index",
    )
    mode = widgets.Dropdown(
        options=("rgb", "bgr"), layout=Layout(width="200px"), description="mode"
    )
    controls = {"index": slider, "mode": mode}
    out_fig = widgets.interactive_output(show_image, controls)
    vbox = widgets.VBox(children=[out_fig, slider, mode])

    return vbox


def show_grayscale_image_series(image_series: ImageSeries, neurodata_vis_spec: dict):
    def show_image(index=0):
        fig, ax = plt.subplots(subplot_kw={"xticks": [], "yticks": []})
        ax.imshow(image_series.data[index].T, cmap="gray", aspect="auto")
        return fig

    slider = widgets.IntSlider(
        value=0,
        min=0,
        max=image_series.data.shape[0] - 1,
        orientation="horizontal",
        continuous_update=False,
        description="index",
    )
    controls = {"index": slider}
    out_fig = widgets.interactive_output(show_image, controls)
    vbox = widgets.VBox(children=[out_fig, slider])

    return vbox


def show_index_series(index_series, neurodata_vis_spec: dict):
    show_timeseries = neurodata_vis_spec[TimeSeries]
    series_widget = show_timeseries(index_series)

    indexed_timeseries = index_series.indexed_timeseries
    image_series_widget = show_image_series(indexed_timeseries, neurodata_vis_spec)

    return widgets.VBox([series_widget, image_series_widget])


def show_grayscale_image(grayscale_image: GrayscaleImage, neurodata_vis_spec=None):
    fig, ax = plt.subplots()
    plt.imshow(grayscale_image.data[:].T, "gray")
    plt.axis("off")

    return fig


def show_rbga_image(rgb_image: RGBImage, neurodata_vis_spec=None):
    fig, ax = plt.subplots()
    plt.imshow(rgb_image.data[:].transpose([1, 0, 2]))
    plt.axis("off")

    return fig



from collections.abc import Iterable
from datetime import datetime
from typing import Union

import pandas as pd
import matplotlib.pyplot as plt

# from ipydatagrid import DataGrid re-add when ipywidgets 8 is supported
from IPython import display
from ipywidgets import widgets
from ipywidgets.widgets.interaction import show_inline_matplotlib_plots

import h5py
from pynwb import ProcessingModule
from pynwb.core import NWBDataInterface, MultiContainerInterface
from pynwb.base import DynamicTable

from nwbwidgets import view

GroupingWidget = Union[widgets.Accordion, widgets.Tab]


def show_fields(node, **kwargs) -> widgets.Widget:
    field_lay = widgets.Layout(
        max_height="40px", max_width="600px", min_height="30px", min_width="130px"
    )
    info = []
    for key, val in node.fields.items():
        lbl_key = widgets.Label(key + ":", layout=field_lay)
        lbl_val = widgets.Label(str(val), layout=field_lay)
        info.append(widgets.HBox(children=[lbl_key, lbl_val]))
    vbox = widgets.VBox(info)
    return vbox


def render_dataframe(dynamic_table: DynamicTable):
    # try:
    #     return DataGrid(dynamic_table.to_dataframe())
    # except:
    out1 = widgets.Output()
    with out1:
        display.display(dynamic_table.to_dataframe())
    return out1


def show_neurodata_base(
    node: NWBDataInterface, neurodata_vis_spec: dict
) -> widgets.Widget:
    """
    Gets a pynwb object and returns a Vertical Box containing textual info and
    an expandable Accordion with it's children.
    """
    field_lay = widgets.Layout(
        max_height="40px", max_width="500px", min_height="30px", min_width="180px"
    )
    info = []  # string data type, exposed as a Text widget
    neuro_data = []  # more complex data types, also with children
    labels = []
    for key, value in node.fields.items():
        if isinstance(value, (str, datetime)):
            lbl_key = widgets.Label(key + ":", layout=field_lay)
            lbl_val = widgets.Label(str(value), layout=field_lay)
            info.append(widgets.HBox(children=[lbl_key, lbl_val]))
        elif key == "related_publications":
            pub_list = []
            for pub in value:
                pub_list.append(
                    widgets.HTML(
                        value="<a href=http://dx.doi.org/"
                        + pub[4:]
                        + ">"
                        + pub
                        + "</a>"
                    )
                )
            lbl_key = widgets.Label(key + ":", layout=field_lay)
            pub_list.insert(0, lbl_key)
            info.append(widgets.HBox(children=pub_list))
        elif key == "experimenter":
            lbl_experimenter = widgets.Label("Experimenter:", layout=field_lay)
            if isinstance(value, (list, tuple)):
                lbl_names = widgets.Label(", ".join(value), layout=field_lay)
            else:
                lbl_names = widgets.Label(value, layout=field_lay)
            hbox_exp = widgets.HBox(children=[lbl_experimenter, lbl_names])
            info.append(hbox_exp)
        elif (isinstance(value, Iterable) and len(value)) or value:
            neuro_data.append(
                view.nwb2widget(value, neurodata_vis_spec=neurodata_vis_spec)
            )
            labels.append(key)
    accordion = widgets.Accordion(children=neuro_data, selected_index=None)
    for i, label in enumerate(labels):
        if (
            hasattr(node.fields[label], "description")
            and node.fields[label].description
        ):
            accordion.set_title(i, label + ": " + node.fields[label].description)
        else:
            accordion.set_title(i, label)
    return widgets.VBox(info + [accordion])


def dict2accordion(d: dict, neurodata_vis_spec: dict, **pass_kwargs) -> widgets.Widget:
    if len(d) == 1:
        return nwb2widget(list(d.values())[0], neurodata_vis_spec=neurodata_vis_spec)
    children = [widgets.HTML("Rendering...") for _ in d]
    accordion = widgets.Accordion(children=children, selected_index=None)
    for i, label in enumerate(d):
        if hasattr(d[label], "description") and d[label].description:
            accordion.set_title(i, label + ": " + d[label].description)
        else:
            accordion.set_title(i, label)

    def on_selected_index(change):
        if change.new is not None and isinstance(
            change.owner.children[change.new], widgets.HTML
        ):
            children[change.new] = nwb2widget(
                list(d.values())[change.new],
                neurodata_vis_spec=neurodata_vis_spec,
                **pass_kwargs
            )
            change.owner.children = children

    accordion.observe(on_selected_index, names="selected_index")

    return accordion


def lazy_tabs(
    in_dict: dict, node, style: GroupingWidget = widgets.Tab
) -> GroupingWidget:
    """Creates a lazy tab object where multiple visualizations can be used for a single node and are generated on the
    fly

    Parameters
    ----------
    in_dict: dict
        keys are labels for tabs and values are functions
    node: NWBDataInterface
        instance of neurodata type to visualize
    style: ipywidgets.Tab or ipywidgets.Accordion, optional
        which way to present the data

    Returns
    -------
    tab: widget

    """
    tabs_spec = list(in_dict.items())

    children = [tabs_spec[0][1](node)] + [
        widgets.HTML("Rendering...") for _ in range(len(tabs_spec) - 1)
    ]
    tab = style(children=children)
    [tab.set_title(i, label) for i, (label, _) in enumerate(tabs_spec)]

    def on_selected_index(change):
        if isinstance(change.owner.children[change.new], widgets.HTML):
            children[change.new] = vis2widget(tabs_spec[change.new][1](node))
            change.owner.children = children

    tab.observe(on_selected_index, names="selected_index")

    return tab


class LazyTab(widgets.Tab):
    """A lazy tab object where multiple visualizations can be used for a single node and are generated on the fly"""

    def __init__(self, func_dict, data):
        """
        Parameters
        ----------
        func_dict: dict
            keys are labels for tabs and values are functions
        data: NWBDataInterface
            instance of neurodata type to visualize
        """

        tabs_spec = list(func_dict.items())
        children = [tabs_spec[0][1](data)] + [
            widgets.HTML("Rendering...") for _ in range(len(tabs_spec) - 1)
        ]

        super().__init__(children=children)

        [self.set_title(i, label) for i, (label, _) in enumerate(tabs_spec)]

        def on_selected_index(change):
            if isinstance(change.owner.children[change.new], widgets.HTML):
                children[change.new] = vis2widget(tabs_spec[change.new][1](data))
                change.owner.children = children

        self.observe(on_selected_index, names="selected_index")


def lazy_show_over_data(
    list_, func_, labels=None, style: GroupingWidget = widgets.Tab
) -> GroupingWidget:
    """
    Apply same function to list of data in lazy tabs or lazy accordion
    Parameters
    ----------
    list_
    func_
    labels: list of str
    style: widgets.Tab or widgets.Accordion

    Returns
    -------
    ipywidgets.Tab or ipywidgets.Accordion
        subtype Tab or Accordion

    """
    children = [vis2widget(func_(list_[0]))] + [
        widgets.HTML("Rendering...") for _ in range(len(list_) - 1)
    ]
    out = style(children=children)
    if labels is not None:
        [out.set_title(i, label) for i, label in enumerate(labels)]

    def on_selected_index(change):
        if change.new is not None and isinstance(
            change.owner.children[change.new], widgets.HTML
        ):
            children[change.new] = vis2widget(func_(list_[change.new]))
            change.owner.children = children

    out.observe(on_selected_index, names="selected_index")

    return out


def nwb2widget(node, neurodata_vis_spec: dict, **pass_kwargs) -> widgets.Widget:
    for ndtype in type(node).__mro__:
        if ndtype in neurodata_vis_spec:
            spec = neurodata_vis_spec[ndtype]
            if isinstance(spec, dict):
                return lazy_tabs(spec, node)
            elif callable(spec):
                return vis2widget(
                    spec(node, neurodata_vis_spec=neurodata_vis_spec, **pass_kwargs)
                )
    out1 = widgets.Output()
    with out1:
        print(node)
    return out1


def vis2widget(vis) -> widgets.Widget:
    if isinstance(vis, widgets.Widget):
        out = vis
    elif isinstance(vis, plt.Figure):
        out = fig2widget(vis)
    elif isinstance(vis, plt.Axes):
        out = fig2widget(vis.get_figure())
    else:
        raise ValueError("unsupported vis type {}".format(type(vis)))

    out.add_class("custom_theme")

    return out


def fig2widget(fig: plt.Figure, **kwargs) -> widgets.Widget:
    out = widgets.Output()
    with out:
        fig.show()
        show_inline_matplotlib_plots()
    return out


def processing_module(
    node: ProcessingModule, neurodata_vis_spec: dict
) -> widgets.Widget:
    return nwb2widget(node.data_interfaces, neurodata_vis_spec=neurodata_vis_spec)


def show_text_fields(
    node, exclude=("comments", "interval"), **kwargs
) -> widgets.Widget:
    info = []
    for key in node.fields:
        if key not in exclude and isinstance(key, (str, float, int)):
            info.append(
                widgets.Text(
                    value=repr(getattr(node, key)), description=key, disabled=True
                )
            )
    return widgets.VBox(info)


def df2accordion(
    df: pd.DataFrame,
    by,
    func,
    style: GroupingWidget = widgets.Accordion,
    detect_single=True,
) -> GroupingWidget:
    """
    Visualize pandas.DataFrame with an ipywidgets.Accordion

    Parameters
    ----------
    df: pandas.DataFrame
    by: str
    func: visualization function
    style: ipywidgets.Tab or ipywidgets.Accordion, optional
    detect_single: bool
        If True, test if the dimension you are grouping by only has 1 unique value. If so, do not form an Accordion.

    Returns
    -------
    ipywigets.Accordion or ipywidgets.Tab


    """
    if detect_single and df[by].nunique() == 1:
        return func(df)
    else:
        labels, idfs = zip(*df.groupby(by))
        return lazy_show_over_data(idfs, func, labels=labels, style=style)


def show_dset(dset: h5py.Dataset, **kwargs):
    return widgets.VBox(children=[show_dict(dict(dset.attrs)), dataset_to_sheet(dset)])


def dataset_to_sheet(dset: h5py.Dataset):
    # if dset.ndim >= 2:
    #     return widgets.HTML(print(dset))
    # return DataGrid(pd.DataFrame(dset[:]))
    return widgets.HTML(print(dset))


def show_dict(in_dict) -> widgets.Widget:
    field_lay = widgets.Layout(
        max_height="40px", max_width="600px", min_height="30px", min_width="130px"
    )
    info = []
    for key, val in in_dict.items():
        lbl_key = widgets.Label(key + ":", layout=field_lay)
        lbl_val = widgets.Label(str(val), layout=field_lay)
        info.append(widgets.HBox(children=[lbl_key, lbl_val]))
    vbox = widgets.VBox(info)
    return vbox


def df_to_hover_text(df: pd.DataFrame):
    return [row_to_hover_text(row) for _, row in df.iterrows()]


def row_to_hover_text(row):
    text_rows = []
    for key, val in list(row.to_dict().items()):
        if isinstance(val, (int, float)) and abs(val) > 1e5:
            text_rows.append("{}: {:.2e}".format(key, val))
        elif isinstance(val, (int, str)):
            text_rows.append("{}: {}".format(key, val))
        elif isinstance(val, float):
            text_rows.append("{}: {:.3f}".format(key, val))
    return "<br>".join(text_rows)


class TimeIntervalsSelector(widgets.VBox):
    InnerWidget = None

    def __init__(self, input_data, **kwargs):
        """
        Creates a TimeInterval controller that controls InnerWidget.

        Parameters
        ----------
        input_data: pynwb object
            Pynwb object (e.g. pynwb.misc.Units) belonging to a nwbfile
            that will be filtered by the TimeIntervalSelector controller.
        """
        super().__init__()
        self.input_data = input_data
        self.kwargs = kwargs
        self.intervals_tables = input_data.get_ancestor("NWBFile").intervals
        self.stimulus_type_dd = widgets.Dropdown(
            options=list(self.intervals_tables.keys()),
            description="stimulus type"
        )
        self.stimulus_type_dd.observe(self.stimulus_type_dd_callback)

        trials = list(self.intervals_tables.values())[0]
        inner_widget = self.InnerWidget(
            units=self.input_data,
            trials=trials,
            **kwargs
        )
        self.children = [self.stimulus_type_dd, inner_widget]

    def stimulus_type_dd_callback(self, change):
        self.children = [self.stimulus_type_dd, widgets.HTML("Rendering...")]
        trials = self.intervals_tables[self.stimulus_type_dd.value]
        inner_widget = self.InnerWidget(
            input_data=self.input_data,
            trials=trials,
            **self.kwargs
        )
        self.children = [self.stimulus_type_dd, inner_widget]


def show_multi_container_interface(
    node: MultiContainerInterface, neurodata_vis_spec=None
):
    if isinstance(node.__clsconf__, dict):
        cls_conf = [node.__clsconf__]
    else:
        cls_conf = node.__clsconf__

    return dict2accordion(
        {x["attr"]: getattr(node, x["attr"]) for x in cls_conf},
        neurodata_vis_spec=neurodata_vis_spec,
    )



from functools import partial

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from ipywidgets import widgets
from plotly.subplots import make_subplots
import plotly.graph_objects as go
import plotly.express as px

from pynwb.icephys import SequentialRecordingsTable

from ndx_icephys_meta.icephys import SweepSequences

from .base import lazy_show_over_data, GroupingWidget
from .timeseries import show_indexed_timeseries_mpl


def show_single_sweep_sequence(
    sweep_sequence, axs=None, title=None, **kwargs
) -> plt.Figure:
    """
    Show a single rep of a single stimulus sequence

    Parameters
    ----------
    sweep_sequence
    axs: [matplotlib.pyplot.Axes, matplotlib.pyplot.Axes], optional
    title: str, optional
    kwargs: dict
        passed to show_indexed_timeseries_mpl

    Returns
    -------

    matplotlib.pyplot.Figure

    """

    nsweeps = len(sweep_sequence)
    if axs is None:
        fig, axs = plt.subplots(2, 1, sharex=True)
    else:
        fig = axs[0].get_figure()
    for i in range(nsweeps):
        start, stop, ts = sweep_sequence["recordings"].iloc[i]["response"].iloc[0][0]
        show_indexed_timeseries_mpl(
            ts,
            istart=start,
            istop=stop,
            ax=axs[0],
            zero_start=True,
            xlabel="",
            title=title,
            **kwargs
        )

        start, stop, ts = sweep_sequence["recordings"].iloc[i]["stimulus"].iloc[0][0]
        show_indexed_timeseries_mpl(
            ts, istart=start, istop=stop, ax=axs[1], zero_start=True, **kwargs
        )
    return fig


def show_sweep_sequence_reps(stim_df: pd.DataFrame, **kwargs) -> plt.Figure:
    """
    Show data from multiple reps of the same stimulus type

    Parameters
    ----------
    stim_df: pandas.DataFrame
    kwargs: dict
        passed to show_single_sweep_sequence

    Returns
    -------
    matplotlib.pyplot.Figure

    """
    nsweeps = len(stim_df["sweeps"])

    if "repetition" in stim_df:
        stim_df = stim_df.sort_values("repetition")
    fig, axs = plt.subplots(
        2, nsweeps, sharex="col", sharey="row", figsize=[6.4 * nsweeps, 4.8]
    )
    if nsweeps == 1:
        axs = np.array([axs]).T
    for i, (sweep, sweep_axs) in enumerate(zip(stim_df["sweeps"], axs.T)):
        if i:
            kwargs.update(ylabel="")
        show_single_sweep_sequence(
            sweep, axs=sweep_axs, title="rep {}".format(i + 1), **kwargs
        )
    return fig


def show_sweep_sequences(
    node: SweepSequences, *args, style: GroupingWidget = widgets.Accordion, **kwargs
) -> GroupingWidget:
    """
    Visualize the sweep sequences table with a lazy accordion of sweep sequence repetitions

    Parameters
    ----------
    node: SweepSequences
    style: widgets.Accordion or widgets.Tabs

    Returns
    -------
    widgets.Accordion or widgets.Tabs

    """
    if "stimulus_type" in node:
        labels, data = zip(
            *[
                (stim_label, stim_df)
                for stim_label, stim_df in node.to_dataframe().groupby("stimulus_type")
            ]
        )
        func_ = show_sweep_sequence_reps
    else:
        data = node["sweeps"]
        labels = None
        func_ = show_single_sweep_sequence
    func_ = partial(func_, **kwargs)
    return lazy_show_over_data(data, func_, labels=labels, style=style)


def show_sequential_recordings(nwbfile, elec_name, sequence_id=0):
    color_wheel = px.colors.qualitative.D3

    stimulus_type = nwbfile.icephys_sequential_recordings[sequence_id]["stimulus_type"].values[0]
    curve_type = "Stimulus-Response curve"

    simultaneous_ids = nwbfile.icephys_sequential_recordings[sequence_id]["simultaneous_recordings"].values[0]
    recordings_ids = np.array([])
    for i in nwbfile.icephys_simultaneous_recordings[simultaneous_ids]["recordings"]:
        recordings_ids = np.append(recordings_ids, i)

    filtered_elec_ids = list()
    for i, row in nwbfile.intracellular_recordings["electrodes"].iterrows():
        if row[0].name == elec_name:
            filtered_elec_ids.append(i)

    filtered_ids = [int(i) for i in np.intersect1d(recordings_ids, filtered_elec_ids)]

    fig = go.FigureWidget(make_subplots(
        rows=2, cols=2,
        specs=[[{}, {"rowspan": 2}], [{}, None]],
        subplot_titles=("", curve_type, stimulus_type)
    ))

    iv_curve_x = list()
    iv_curve_y = list()
    ii = 0
    for i, row in nwbfile.intracellular_recordings.to_dataframe().loc[filtered_ids].iterrows():
        if ii == 0:
            response_unit = row.responses.response.timeseries.unit
            stimulus_unit = row.stimuli.stimulus.timeseries.unit

        response_conversion = row.responses.response.timeseries.conversion
        response_data = np.array(row.responses.response.timeseries.data[:]) * response_conversion
        response_rate = row.responses.response.timeseries.rate
        response_x = np.arange(len(response_data)) / response_rate

        if row.stimuli.stimulus.timeseries:
            stimulus_data = row.stimuli.stimulus.timeseries.data[:]
            stimulus_rate = row.stimuli.stimulus.timeseries.rate
        else:
            stimulus_data = np.zeros(len(response_data))
            stimulus_rate = response_rate
        stimulus_x = np.arange(len(stimulus_data)) / stimulus_rate

        # I-V curve
        if (max(stimulus_data) - stimulus_data[0]) > 0:
            iv_curve_x_point = max(stimulus_data)
        else:
            iv_curve_x_point = min(stimulus_data)
        iv_curve_x.append(iv_curve_x_point)

        abs_response_data = np.absolute(response_data - response_data[0])
        ind = np.argmax(abs_response_data)
        iv_curve_y.append(response_data[ind])

        fig.add_trace(
            go.Scatter(
                x=response_x,
                y=response_data,
                legendgroup=f"{int(ii)}",
                name=f"Sweep {int(ii)}",
                marker=dict(color=color_wheel[int(ii%10)])
            ),
            row=1,
            col=1
        )

        fig.add_trace(
            go.Scatter(
                x=stimulus_x,
                y=stimulus_data,
                legendgroup=f"{int(ii)}",
                showlegend=False,
                marker=dict(color=color_wheel[int(ii%10)])
            ),
            row=2,
            col=1
        )

        ii += 1

    fig.add_trace(
        go.Scatter(
            x=iv_curve_x,
            y=iv_curve_y,
            showlegend=False,
            mode='lines',
            line=dict(color='black', width=2)
        ),
        row=1,
        col=2
    )

    for ii in range(len(iv_curve_x)):
        fig.add_trace(
            go.Scatter(
                x=[iv_curve_x[ii]],
                y=[iv_curve_y[ii]],
                showlegend=False,
                legendgroup=f"{int(ii)}",
                marker=dict(
                    color=color_wheel[int(ii%10)],
                    size=10
                )
            ),
            row=1,
            col=2
        )

    fig.update_layout(
        height=600,
        width=1200,
        legend=dict(
            orientation="h",
            yanchor="bottom",
            y=1.1,
            xanchor="right",
            x=1
        )
    )
    fig.update_xaxes(showgrid=False, row=1, col=1)
    fig.update_xaxes(title_text="time [s]", showgrid=False, row=2, col=1)
    fig.update_xaxes(title_text=f"stimuli [{stimulus_unit}]",showgrid=False, row=1, col=2)
    fig.update_xaxes(showline=False, zeroline=True, zerolinewidth=1, zerolinecolor='black')

    fig.update_yaxes(title_text=f"response [{response_unit}]",showgrid=False, row=1, col=1)
    fig.update_yaxes(title_text=f"stimuli [{stimulus_unit}]",showgrid=False, row=2, col=1)
    fig.update_yaxes(title_text=f"response [{response_unit}]",showgrid=False, row=1, col=2)
    fig.update_yaxes(showline=False, zeroline=True, zerolinewidth=1, zerolinecolor='black')

    return fig


class IVCurveWidget(widgets.VBox):
    def __init__(
        self,
        sequential_recordings_table: SequentialRecordingsTable,
        neurodata_vis_spec=None,
        **kwargs
    ):
        super().__init__()

        self.table = sequential_recordings_table

        # Electrodes
        elec_options = [(n, i) for i, n in enumerate(self.table.get_ancestor().ic_electrodes.keys())]
        self.electrode_name = elec_options[0][0]
        dropdown_elec = widgets.Dropdown(
            options=elec_options,
            value=0,
            description='Electrode:',
        )
        dropdown_elec.observe(self.update_electrode)

        # Stimuli
        self.stimuli_index = 0
        dropdown_stim = widgets.Dropdown(
            options=[(v, i) for i, v in enumerate(list(self.table.stimulus_type[:]))],
            value=0,
            description='Stimulus:',
        )
        dropdown_stim.observe(self.update_stimulus)

        self.iv_curve_controller = widgets.HBox([
            dropdown_elec,
            dropdown_stim
        ])

        self.update_figure()

        self.children = [self.iv_curve_controller, self.fig]


    def update_electrode(self, change):
        if change["name"] == "label":
            self.electrode_name = change["new"]
            self.update_figure()


    def update_stimulus(self, change):
        if change["name"] == "value":
            self.stimuli_index = change["new"]
            self.update_figure()


    def update_figure(self):
        self.fig = show_sequential_recordings(
            nwbfile=self.table.get_ancestor(),
            elec_name=self.electrode_name,
            sequence_id=self.stimuli_index
        )
        self.children = [self.iv_curve_controller, self.fig]



from datetime import datetime
from dateutil.tz import tzlocal

import numpy as np
import matplotlib.pyplot as plt

from ipywidgets import widgets
import plotly.graph_objects as go

from pynwb import NWBFile
from pynwb.base import TimeSeries
from pynwb.device import Device
from pynwb.icephys import IntracellularElectrode, VoltageClampStimulusSeries, VoltageClampSeries

from ndx_icephys_meta.icephys import Sweeps, IntracellularRecordings

from nwbwidgets.icephys import show_single_sweep_sequence, IVCurveWidget


def test_show_single_sweep_sequence():
    device = Device(name="Axon Patch-Clamp")
    electrode = IntracellularElectrode(
        name="Patch Clamp", device=device, description="whole-cell"
    )

    stimulus_data = np.random.rand(160, 2)
    stimulus = TimeSeries(
        name="test_timeseries",
        data=stimulus_data,
        unit="m",
        starting_time=0.0,
        rate=1.0,
    )
    response_data = np.random.rand(160, 2)
    response = TimeSeries(
        name="test_timeseries",
        data=response_data,
        unit="m",
        starting_time=0.0,
        rate=1.0,
    )

    icr = IntracellularRecordings()
    icr.add_recording(
        electrode=electrode,
        stimulus_start_index=0,
        stimulus_index_count=100,
        stimulus=stimulus,
        response_start_index=0,
        response_index_count=100,
        response=response,
    )

    sweeps_table = Sweeps(intracellular_recordings_table=icr)
    assert isinstance(show_single_sweep_sequence(sweeps_table), plt.Figure)


def test_iv_curve():
    # Create an ICEphysFile
    ex_nwbfile = NWBFile(
        session_description='my first recording',
        identifier='EXAMPLE_ID',
        session_start_time=datetime.now(tzlocal())
    )

    # Add a device
    ex_device = ex_nwbfile.create_device(name='Heka ITC-1600')

    # Add an intracellular electrode
    ex_electrode = ex_nwbfile.create_icephys_electrode(
        name="elec0",
        description='a mock intracellular electrode',
        device=ex_device
    )

    # Create an ic-ephys stimulus
    ex_stimulus = VoltageClampStimulusSeries(
        name="stimulus",
        data=[1, 2, 3, 4, 5],
        starting_time=123.6,
        rate=10e3,
        electrode=ex_electrode,
        gain=0.02
    )

    # Create an ic-response
    ex_response = VoltageClampSeries(
        name='response',
        data=[0.1, 0.2, 0.3, 0.4, 0.5],
        conversion=1e-12,
        resolution=np.nan,
        starting_time=123.6,
        rate=20e3,
        electrode=ex_electrode,
        gain=0.02,
        capacitance_slow=100e-12,
        resistance_comp_correction=70.0
    )

    # (A) Add an intracellular recording to the file
    ex_ir_index = ex_nwbfile.add_intracellular_recording(
        electrode=ex_electrode,
        stimulus=ex_stimulus,
        response=ex_response
    )

    # (B) Add a list of sweeps to the simultaneous recordings table
    ex_sweep_index = ex_nwbfile.add_icephys_simultaneous_recording(recordings=[ex_ir_index, ])

    # (C) Add a list of simultaneous recordings table indices as a sequential recording
    ex_sequence_index = ex_nwbfile.add_icephys_sequential_recording(
        simultaneous_recordings=[ex_sweep_index, ],
        stimulus_type='square'
    )

    # (D) Add a list of sequential recordings table indices as a repetition
    run_index = ex_nwbfile.add_icephys_repetition(sequential_recordings=[ex_sequence_index, ])

    # (E) Add a list of repetition table indices as a experimental condition
    ex_nwbfile.add_icephys_experimental_condition(repetitions=[run_index, ])

    # Make widget
    w = IVCurveWidget(sequential_recordings_table=ex_nwbfile.icephys_sequential_recordings)
    assert isinstance(w, widgets.VBox)
    assert isinstance(w.fig, go.FigureWidget)



import unittest

import numpy as np
import matplotlib.pyplot as plt

from ipywidgets import widgets
import plotly.graph_objects as go

from pynwb import TimeSeries
from pynwb.epoch import TimeIntervals
from pynwb.behavior import SpatialSeries

from nwbwidgets.timeseries import (
    BaseGroupedTraceWidget,
    show_ts_fields,
    show_timeseries,
    plot_traces,
    show_indexed_timeseries_mpl,
    AlignMultiTraceTimeSeriesByTrialsConstant,
    AlignMultiTraceTimeSeriesByTrialsVariable,
    SingleTracePlotlyWidget,
    SeparateTracesPlotlyWidget,
    get_timeseries_tt,
    show_indexed_timeseries_plotly,
    show_timeseries_mpl,
)


def test_timeseries_widget():
    ts = TimeSeries(
        name="name",
        description="no description",
        data=np.array([[1.0, 2.0, 3.0, 4.0], [11.0, 12.0, 13.0, 14.0]]),
        rate=100.0,
        unit="m",
    )

    BaseGroupedTraceWidget(ts)


def test_show_timeseries_mpl():

    ts = TimeSeries(
        name="name",
        description="no description",
        data=np.array([[1.0, 2.0, 3.0, 4.0], [11.0, 12.0, 13.0, 14.0]]),
        rate=100.0,
        unit="m",
    )

    show_timeseries_mpl(ts)
    show_timeseries_mpl(ts, time_window=(0.0, 1.0))


class TestTracesPlotlyWidget(unittest.TestCase):
    def setUp(self):
        data = np.random.rand(160, 3)
        self.ts_multi = SpatialSeries(
            name="test_timeseries_multi",
            data=data,
            reference_frame="lowerleft",
            starting_time=0.0,
            rate=1.0,
        )
        self.ts_single = TimeSeries(
            name="test_timeseries_single",
            data=data[:, 0],
            unit="m",
            starting_time=0.0,
            rate=1.0,
        )
        # Create 'intermittent' timeseries by defining timestamps with a gap from [10, 20]
        self.ts_single_intermittent = TimeSeries(
            name="test_timeseries_single_intermittent",
            data=data[:, 0],
            timestamps=np.hstack([np.arange(0, 10, 10 / data.shape[0] * 2),
                                  np.arange(20, 30, 10 / data.shape[0] * 2)]),
            unit='m',
        )
        self.ts_multi_intermittent = SpatialSeries(
            name="test_timeseries_multi_intermittent",
            data=data,
            timestamps=np.hstack([np.arange(0, 10, 10 / data.shape[0] * 2),
                                  np.arange(20, 30, 10 / data.shape[0] * 2)]),
            reference_frame="lowerleft",
        )


    def test_single_trace_widget(self):
        single_wd = SingleTracePlotlyWidget(timeseries=self.ts_single)
        tt = get_timeseries_tt(self.ts_single)
        single_wd.controls["time_window"].value = [
            tt[int(len(tt) * 0.2)],
            tt[int(len(tt) * 0.4)],
        ]

    def test_single_trace_widget_null_data(self):
        single_wd_nd = SingleTracePlotlyWidget(timeseries=self.ts_single_intermittent)
        single_wd_nd.controls["time_window"].value = [12, 14]

    def test_separate_traces_widget(self):
        sep_wd = SeparateTracesPlotlyWidget(timeseries=self.ts_multi)
        tt = get_timeseries_tt(self.ts_multi)
        sep_wd.controls["time_window"].value = [
            tt[int(len(tt) * 0.2)],
            tt[int(len(tt) * 0.4)],
        ]

    def test_separate_traces_widget_null_data(self):
        single_wd_nd = SingleTracePlotlyWidget(timeseries=self.ts_multi_intermittent)
        single_wd_nd.controls["time_window"].value = [12, 14]


class TestIndexTimeSeriesPlotly(unittest.TestCase):
    def setUp(self):
        data = np.random.rand(160, 3)
        self.ts = TimeSeries(
            name="name",
            description="no description",
            data=data,
            starting_time=0.0,
            rate=100.0,
            unit="m",
        )
        self.ts_single = TimeSeries(
            name="test_timeseries",
            data=data[:, 0],
            unit="m",
            starting_time=0.0,
            rate=100.0,
        )
        self.tt = get_timeseries_tt(self.ts)

    def test_no_args(self):
        fig_out = show_indexed_timeseries_plotly(timeseries=self.ts)
        fig_out_single = show_indexed_timeseries_plotly(timeseries=self.ts_single)
        assert isinstance(fig_out, go.FigureWidget)
        assert isinstance(fig_out_single, go.FigureWidget)
        assert len(fig_out.data) == 3
        assert len(fig_out_single.data) == 1
        assert np.allclose(fig_out.data[0].x, self.tt)
        assert np.allclose(fig_out_single.data[0].x, self.tt)
        assert np.allclose(fig_out.data[0].y, self.ts.data[:, 0])
        assert np.allclose(fig_out.data[1].y, self.ts.data[:, 1])
        assert np.allclose(fig_out.data[2].y, self.ts.data[:, 2])

    def test_value_errors(self):
        time_window = [
            self.tt[int(len(self.tt) * 0.2)],
            self.tt[int(len(self.tt) * 0.4)],
        ]
        self.assertRaises(
            ValueError,
            show_indexed_timeseries_plotly,
            timeseries=self.ts_single,
            istart=3,
            time_window=time_window,
        )
        self.assertRaises(
            ValueError,
            show_indexed_timeseries_plotly,
            timeseries=self.ts_single,
            trace_range=[2, 5],
        )


class ShowTimeSeriesTestCase(unittest.TestCase):
    def setUp(self):
        data = np.random.rand(160, 3)
        self.ts = TimeSeries(
            name="test_timeseries", data=data, unit="m", starting_time=0.0, rate=1.0
        )

    def test_show_ts_fields(self):
        assert isinstance(show_ts_fields(self.ts), widgets.Widget)

    def test_show_timeseries(self):
        assert isinstance(show_timeseries(self.ts, istart=5, istop=56), widgets.Widget)

    def test_show_indexed_timeseries_mpl(self):
        ax = show_indexed_timeseries_mpl(
            self.ts, zero_start=True, title="Test show_indexed_timeseries_mpl"
        )
        assert isinstance(ax, plt.Subplot)


class PlotTracesTestCase(unittest.TestCase):
    def setUp(self):
        self.data = np.random.rand(160, 3)

    def test_plot_traces(self):
        ts = TimeSeries(
            name="test_timeseries",
            data=self.data,
            unit="m",
            starting_time=0.0,
            rate=20.0,
        )
        plot_traces(ts)

    def test_plot_traces_fix(self):
        ts = TimeSeries(
            name="test_timeseries",
            data=self.data.T,
            unit="m",
            starting_time=0.0,
            rate=20.0,
        )
        plot_traces(ts)


class TestAlignMultiTraceTimeSeriesByTrials(unittest.TestCase):
    def setUp(self):
        data = np.random.rand(100, 10)
        timestamps = [0.0]
        for _ in range(data.shape[0]):
            timestamps.append(timestamps[-1] + 0.75 + 0.25 * np.random.rand())
        self.ts_rate = TimeSeries(
            name="test_timeseries_rate",
            data=data,
            unit="m",
            starting_time=0.0,
            rate=1.0,
        )
        self.ts_timestamps = TimeSeries(
            name="test_timeseries_timestamps",
            data=data,
            unit="m",
            timestamps=np.array(timestamps),
        )
        self.time_intervals = TimeIntervals(name="Test Time Interval")
        n_intervals = 10
        for start_time in np.linspace(0, 75, n_intervals + 1):
            if start_time < 75:
                stt = start_time + np.random.rand()
                spt = stt + 7 - np.random.rand()
                self.time_intervals.add_interval(start_time=stt, stop_time=spt)
        self.time_intervals.add_column(
            name="temp", description="desc", data=np.random.randint(2, size=n_intervals)
        )
        self.time_intervals.add_column(
            name="temp2",
            description="desc",
            data=np.random.randint(10, size=n_intervals),
        )

    def test_align_by_timestamps(self):
        amt = AlignMultiTraceTimeSeriesByTrialsVariable(
            time_series=self.ts_timestamps, trials=self.time_intervals
        )
        gas = amt.controls['gas']
        gas.group_dd.value = list(gas.categorical_columns.keys())[0]
        gas.group_sm.value = (gas.group_sm.options[0],)
        fig = amt.children[-1]
        assert len(fig.data)==len(gas.group_sm.value)

    def test_align_by_rate(self):
        amt = AlignMultiTraceTimeSeriesByTrialsConstant(
            time_series=self.ts_rate, trials=self.time_intervals
        )
        gas = amt.controls['gas']
        gas.group_dd.value = list(gas.categorical_columns)[0]
        gas.group_sm.value = (gas.group_sm.options[0],)
        fig = amt.children[-1]
        assert len(fig.data) == len(gas.group_sm.value)



from abc import abstractmethod

from ipywidgets import widgets, Layout, ValueWidget, link, HBox
from ipywidgets.widgets.widget_description import DescriptionWidget


class WindowController(HBox, ValueWidget, DescriptionWidget):
    def __init__(
        self,
        vmin,
        vmax,
        start_value=(None, None),
        description="window (s)",
        orientation="horizontal",
        **kwargs
    ):

        if orientation not in ("horizontal", "vertical"):
            ValueError("Unrecognized orientation: {}".format(orientation))

        super().__init__()
        self.vmin = vmin
        self.vmax = vmax
        self.start_value = start_value
        self.description = description
        self.orientation = orientation

        if self.orientation == "horizontal":
            self.to_start_button = widgets.Button(
                description="◀◀", layout=Layout(width="65px")
            )
            self.backwards_button = widgets.Button(
                description="◀", layout=Layout(width="40px")
            )
            self.forward_button = widgets.Button(
                description="▶", layout=Layout(width="40px")
            )
            self.to_end_button = widgets.Button(
                description="▶▶", layout=Layout(width="65px")
            )
        else:  # vertical
            self.to_end_button = widgets.Button(
                description="▲▲", layout=Layout(width="50px")
            )
            self.forward_button = widgets.Button(
                description="▲", layout=Layout(width="50px")
            )
            self.backwards_button = widgets.Button(
                description="▼", layout=Layout(width="50px")
            )
            self.to_start_button = widgets.Button(
                description="▼▼", layout=Layout(width="50px")
            )

        self.to_start_button.on_click(self.move_start)
        self.backwards_button.on_click(self.move_down)
        self.forward_button.on_click(self.move_up)
        self.to_end_button.on_click(self.move_end)

    # @abstractmethod
    # @property
    # def value(self):
    #    """Must be window in seconds"""
    #    pass

    @abstractmethod
    def get_children(self):
        pass

    @abstractmethod
    def move_up(self, change):
        pass

    @abstractmethod
    def move_down(self, change):
        pass

    @abstractmethod
    def move_start(self, change):
        pass

    @abstractmethod
    def move_end(self, change):
        pass


class RangeController(WindowController):
    def __init__(
        self,
        vmin,
        vmax,
        start_value=None,
        dtype="float",
        description="time window (s)",
        orientation="horizontal",
        **kwargs
    ):
        super().__init__(vmin, vmax, start_value, description, orientation, **kwargs)

        self.dtype = dtype
        self.slider = self.make_range_slider(description=description, **kwargs)

        link((self.slider, "value"), (self, "value"))
        link((self.slider, "description"), (self, "description"))

        self.children = self.get_children()

    def make_range_slider(self, **kwargs):
        """

        Parameters
        ----------
        kwargs: passed into RangeSlider constructor

        Returns
        -------

        """

        slider_kwargs = dict(
            value=self.start_value,
            min=self.vmin,
            max=self.vmax,
            continuous_update=False,
            readout=True,
            style={"description_width": "initial"},
            orientation=self.orientation,
        )

        if self.dtype == "float":
            slider_kwargs.update(
                readout_format=".1f",
                step=0.1,
                description="time window (s)",
                layout=Layout(width="100%"),
            )
            slider_kwargs.update(kwargs)
            return widgets.FloatRangeSlider(**slider_kwargs)
        elif self.dtype == "int":
            slider_kwargs.update(
                description="unit window", layout=Layout(height="100%")
            )
            slider_kwargs.update(kwargs)
            return widgets.IntRangeSlider(**slider_kwargs)
        else:
            raise ValueError("Unrecognized dtype: {}".format(self.dtype))

    def move_up(self, change):
        value_range = self.value[1] - self.value[0]
        if self.value[1] + value_range < self.vmax:
            self.value = (self.value[0] + value_range, self.value[1] + value_range)
        else:
            self.move_end(change)

    def move_down(self, change):
        value_range = self.value[1] - self.value[0]
        if self.value[0] - value_range > self.vmin:
            self.value = (self.value[0] - value_range, self.value[1] - value_range)
        else:
            self.move_start(change)

    def move_start(self, change):
        value_range = self.value[1] - self.value[0]
        self.value = (self.vmin, self.vmin + value_range)

    def move_end(self, change):
        value_range = self.value[1] - self.value[0]
        self.value = (self.vmax - value_range, self.vmax)

    def get_children(self):

        if self.orientation == "horizontal":
            return [
                self.slider,
                self.to_start_button,
                self.forward_button,
                self.backwards_button,
                self.to_end_button,
            ]
        else:
            return [
                widgets.VBox(
                    [
                        self.slider,
                        self.to_end_button,
                        self.forward_button,
                        self.backwards_button,
                        self.to_start_button,
                    ],
                    layout=widgets.Layout(
                        display="flex", flex_flow="column", align_items="center"
                    ),
                )
            ]


class StartAndDurationController(WindowController):
    """
    Can be used in place of the RangeController.
    """

    DEFAULT_DURATION = 5

    def __init__(
        self, tmax, tmin=0, start_value=None, description="start (s)", **kwargs
    ):
        """

        Parameters
        ----------
        tmax: float
            in seconds
        tmin: float
            in seconds
        start_value: (float, float)
            start and stop in seconds
        description: str
        kwargs: dict
        """

        if tmin > tmax:
            raise ValueError(
                "tmax and tmin were probably entered in the wrong order. tmax should be first"
            )

        super().__init__(tmin, tmax, start_value, description, **kwargs)

        self.slider = widgets.FloatSlider(
            value=start_value,
            min=tmin,
            max=tmax,
            step=0.01,
            description=description,
            continuous_update=False,
            orientation="horizontal",
            readout=True,
            readout_format=".2f",
            style={"description_width": "initial"},
            layout=Layout(width="100%", min_width="250px"),
        )

        link((self.slider, "description"), (self, "description"))

        if start_value is None:
            duration = min(self.DEFAULT_DURATION, tmax - tmin)
        else:
            duration = start_value[1] - start_value[0]

        self.duration = widgets.BoundedFloatText(
            value=duration,
            min=0,
            max=tmax - tmin,
            step=0.1,
            description="duration (s):",
            style={"description_width": "initial"},
            layout=Layout(max_width="200px"),
        )

        self.value = (self.slider.value, self.slider.value + self.duration.value)

        self.slider.observe(self.monitor_slider)
        self.duration.observe(self.monitor_duration)

        self.children = self.get_children()

    def monitor_slider(self, change):
        if "new" in change:
            if isinstance(change["new"], dict):
                if "value" in change["new"]:
                    value = change["new"]["value"]
                else:
                    return
            else:
                value = change["new"]
        if self.slider.value + self.duration.value > self.vmax:
            self.slider.value = self.vmax - self.duration.value
        else:
            self.value = (value, value + self.duration.value)

    def monitor_duration(self, change):
        if "new" in change:
            if isinstance(change["new"], dict):
                if "value" in change["new"]:
                    value = change["new"]["value"]
                    if self.slider.value + value > self.vmax:
                        self.slider.value = self.vmax - value
                    self.value = (self.slider.value, self.slider.value + value)

    def move_up(self, change):
        if self.slider.value + 2 * self.duration.value < self.vmax:
            self.slider.value += self.duration.value
        else:
            self.move_end(change)

    def move_down(self, change):
        if self.slider.value - self.duration.value > self.vmin:
            self.slider.value -= self.duration.value
        else:
            self.move_start(change)

    def move_start(self, change):
        self.slider.value = self.vmin

    def move_end(self, change):
        self.slider.value = self.vmax - self.duration.value

    def get_children(self):

        if self.orientation == "horizontal":
            return [
                self.slider,
                self.duration,
                # self.to_start_button,
                self.backwards_button,
                self.forward_button,
                # self.to_end_button
            ]
        else:
            return [
                widgets.VBox(
                    [
                        self.slider,
                        self.duration,
                        # self.to_end_button,
                        self.forward_button,
                        self.backwards_button,
                        # self.to_start_button,
                    ],
                    layout=widgets.Layout(
                        display="flex", flex_flow="column", align_items="center"
                    ),
                )
            ]



from collections import OrderedDict

import h5py
import hdmf
import pynwb

from ipywidgets import widgets

import zarr
import ndx_grayscalevolume
from ndx_icephys_meta.icephys import SweepSequences
from ndx_spectrum import Spectrum

from .dynamictablesummary import DynamicTableSummaryWidget

from nwbwidgets import (
    behavior,
    misc,
    base,
    ecephys,
    image,
    ophys,
    icephys,
    timeseries,
    file,
    spectrum,
)


# def show_dynamic_table(node: DynamicTable, **kwargs):
def show_dynamic_table(node, **kwargs) -> widgets.Widget:
    if node.name == "electrodes":
        return ecephys.show_electrodes(node)
    return base.render_dataframe(node)


default_neurodata_vis_spec = {
    pynwb.NWBFile: file.show_nwbfile,
    SweepSequences: icephys.show_sweep_sequences,
    pynwb.behavior.BehavioralEvents: behavior.show_behavioral_events,
    pynwb.misc.Units: OrderedDict(
        {
            "Summary": DynamicTableSummaryWidget,
            "Session Raster": misc.RasterWidget,
            "Grouped PSTH": misc.PSTHWidget,
            "Raster Grid": misc.RasterGridWidget,
            "Tuning Curves": misc.TuningCurveWidget,
            "Combined": misc.TuningCurveExtendedWidget,
            "table": show_dynamic_table,
        }
    ),
    pynwb.misc.DecompositionSeries: misc.show_decomposition_series,
    pynwb.file.Subject: base.show_fields,
    pynwb.ecephys.SpikeEventSeries: ecephys.show_spike_event_series,
    pynwb.ophys.ImageSegmentation: ophys.show_image_segmentation,
    pynwb.ophys.TwoPhotonSeries: ophys.TwoPhotonSeriesWidget,
    ndx_grayscalevolume.GrayscaleVolume: ophys.show_grayscale_volume,
    pynwb.ophys.PlaneSegmentation: ophys.route_plane_segmentation,
    pynwb.ophys.DfOverF: ophys.show_df_over_f,
    pynwb.ophys.RoiResponseSeries: ophys.RoiResponseSeriesWidget,
    pynwb.misc.AnnotationSeries: OrderedDict(
        {
            "text": base.show_text_fields,
            "times": misc.show_annotations
        }
    ),
    pynwb.core.LabelledDict: base.dict2accordion,
    pynwb.ProcessingModule: base.processing_module,
    hdmf.common.DynamicTable:
        {
        "Summary": DynamicTableSummaryWidget,
        "table": show_dynamic_table,
        },
    pynwb.ecephys.ElectricalSeries: ecephys.ElectricalSeriesWidget,
    pynwb.behavior.SpatialSeries: behavior.route_spatial_series,
    pynwb.image.GrayscaleImage: image.show_grayscale_image,
    pynwb.image.RGBImage: image.show_rbga_image,
    pynwb.image.RGBAImage: image.show_rbga_image,
    pynwb.base.Image: image.show_rbga_image,
    pynwb.image.ImageSeries: image.ImageSeriesWidget,
    pynwb.image.IndexSeries: image.show_index_series,
    pynwb.TimeSeries: timeseries.show_timeseries,
    pynwb.core.MultiContainerInterface: base.show_multi_container_interface,
    pynwb.core.NWBContainer: base.show_neurodata_base,
    pynwb.core.NWBDataInterface: base.show_neurodata_base,
    h5py.Dataset: base.show_dset,
    zarr.core.Array: base.show_dset,
    Spectrum: spectrum.show_spectrum,
    pynwb.icephys.SequentialRecordingsTable: {
        "Summary": DynamicTableSummaryWidget,
        "table": show_dynamic_table,
        "I-V Analysis": icephys.IVCurveWidget
    }
}


def nwb2widget(node, neurodata_vis_spec=default_neurodata_vis_spec):
    return base.nwb2widget(node, neurodata_vis_spec)



import functools
from bisect import bisect
from abc import abstractmethod

import scipy
import numpy as np
import matplotlib.pyplot as plt

from ipywidgets import widgets, fixed, Layout
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from plotly.colors import DEFAULT_PLOTLY_COLORS

from pynwb import TimeSeries
from pynwb.epoch import TimeIntervals

from .controllers import (
    StartAndDurationController,
    GroupAndSortController,
    RangeController,
    ProgressBar
)
from .utils.plotly import multi_trace
from .utils.timeseries import (
    get_timeseries_tt,
    get_timeseries_maxt,
    get_timeseries_mint,
    timeseries_time_to_ind,
    get_timeseries_in_units,
)
from .utils.widgets import interactive_output, set_plotly_callbacks
from .controllers.misc import make_trial_event_controller


color_wheel = plt.rcParams["axes.prop_cycle"].by_key()["color"]


def show_ts_fields(node):
    info = []
    for key in ("description", "unit", "resolution", "conversion"):
        info.append(
            widgets.Text(value=repr(getattr(node, key)), description=key, disabled=True)
        )
    return widgets.VBox(info)


def show_timeseries_mpl(
    time_series: TimeSeries,
    time_window=None,
    ax=None,
    zero_start=False,
    xlabel=None,
    ylabel=None,
    title=None,
    figsize=None,
    **kwargs
):
    """

    Parameters
    ----------
    time_series: TimeSeries
    time_window: [int int]
    ax: plt.Axes
    zero_start: bool
    xlabel: str
    ylabel: str
    title: str
    figsize: tuple, optional
    kwargs

    Returns
    -------

    """
    if time_window is not None:
        istart = timeseries_time_to_ind(time_series, time_window[0])
        istop = timeseries_time_to_ind(time_series, time_window[1])
    else:
        istart = 0
        istop = None

    return show_indexed_timeseries_mpl(
        time_series,
        istart=istart,
        istop=istop,
        ax=ax,
        zero_start=zero_start,
        xlabel=xlabel,
        ylabel=ylabel,
        title=title,
        figsize=figsize,
        **kwargs,
    )


def show_indexed_timeseries_mpl(
    node: TimeSeries,
    istart=0,
    istop=None,
    ax=None,
    zero_start=False,
    xlabel="time (s)",
    ylabel=None,
    title=None,
    figsize=None,
    neurodata_vis_spec=None,
    **kwargs
):
    if ylabel is None and node.unit:
        ylabel = node.unit

    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)

    tt = get_timeseries_tt(node, istart=istart, istop=istop)
    if zero_start:
        tt = tt - tt[0]
    data, unit = get_timeseries_in_units(node, istart=istart, istop=istop)

    ax.plot(tt, data, **kwargs)
    ax.set_xlabel(xlabel)
    if ylabel is not None:
        ax.set_ylabel(ylabel)
    if title is not None:
        ax.set_title(title)

    ax.autoscale(enable=True, axis="x", tight=True)

    return ax


def show_indexed_timeseries_plotly(
    timeseries: TimeSeries,
    istart: int = 0,
    istop: int = None,
    time_window: list = None,
    trace_range: list = None,
    offsets=None,
    fig: go.FigureWidget = None,
    col=None,
    row=None,
    zero_start=False,
    scatter_kwargs: dict = None,
    figure_kwargs: dict = None,
):
    if istart != 0 or istop is not None:
        if time_window is not None:
            raise ValueError("input either time window or istart/stop but not both")
        if not (
            0 <= istart < timeseries.data.shape[0]
            and (istop is None or 0 < istop <= timeseries.data.shape[0])
        ):
            raise ValueError("enter correct istart/stop values")
        t_istart = istart
        t_istop = istop
    elif time_window is not None:
        t_istart = timeseries_time_to_ind(timeseries, time_window[0])
        t_istop = timeseries_time_to_ind(timeseries, time_window[1])
    else:
        t_istart = istart
        t_istop = istop
    tt = get_timeseries_tt(timeseries, istart=t_istart, istop=t_istop)
    data, unit = get_timeseries_in_units(timeseries, istart=t_istart, istop=t_istop)
    if len(data.shape) == 1:
        data = data[:, np.newaxis]
    if trace_range is not None:
        if not (
            0 <= trace_range[0] < data.shape[1] and 0 < trace_range[1] <= data.shape[1]
        ):
            raise ValueError("enter correct trace range")
        trace_istart = trace_range[0]
        trace_istop = trace_range[1]
    else:
        trace_istart = 0
        trace_istop = data.shape[1]
    if offsets is None:
        offsets = np.zeros(trace_istop - trace_istart)
    if zero_start:
        tt = tt - tt[0]
    scatter_kwargs = dict() if scatter_kwargs is None else scatter_kwargs
    if fig is None:
        fig = go.FigureWidget(make_subplots(rows=1, cols=1))
    row = 1 if row is None else row
    col = 1 if col is None else col
    for i, trace_id in enumerate(range(trace_istart, trace_istop)):
        fig.add_trace(
            go.Scattergl(
                x=tt, y=data[:, trace_id] + offsets[i], mode="lines", **scatter_kwargs
            ),
            row=row,
            col=col,
        )
    input_figure_kwargs = dict(
        xaxis=dict(title_text="time (s)", range=[tt[0], tt[-1]]),
        yaxis=dict(title_text=unit if unit is not None else None),
        title=timeseries.name,
    )
    if figure_kwargs is None:
        figure_kwargs = dict()
    input_figure_kwargs.update(figure_kwargs)
    fig.update_xaxes(input_figure_kwargs.pop("xaxis"), row=row, col=col)
    fig.update_yaxes(input_figure_kwargs.pop("yaxis"), row=row, col=col)
    fig.update_layout(**input_figure_kwargs)
    return fig


def plot_traces(
    timeseries: TimeSeries,
    time_window=None,
    trace_window=None,
    title: str = None,
    ylabel: str = "traces",
    **kwargs
):
    """

    Parameters
    ----------
    timeseries: TimeSeries
    time_window: [float, float], optional
        Start time and end time in seconds.
    trace_window: [int int], optional
        Index range of traces to view
    title: str, optional
    ylabel: str, optional

    Returns
    -------

    """

    if time_window is None:
        t_ind_start = 0
        t_ind_stop = None
    else:
        t_ind_start = timeseries_time_to_ind(timeseries, time_window[0])
        t_ind_stop = timeseries_time_to_ind(timeseries, time_window[1])

    if trace_window is None:
        trace_window = [0, timeseries.data.shape[1]]
    tt = get_timeseries_tt(timeseries, t_ind_start, t_ind_stop)
    if timeseries.data.shape[1] == len(tt):  # fix of orientation is incorrect
        mini_data = timeseries.data[
            trace_window[0] : trace_window[1], t_ind_start:t_ind_stop
        ].T
    else:
        mini_data = timeseries.data[
            t_ind_start:t_ind_stop, trace_window[0] : trace_window[1]
        ]

    gap = np.median(np.nanstd(mini_data, axis=0)) * 20
    offsets = np.arange(trace_window[1] - trace_window[0]) * gap

    fig, ax = plt.subplots()
    ax.figure.set_size_inches(12, 6)
    ax.plot(tt, mini_data + offsets, **kwargs)
    ax.set_xlabel("time (s)")
    if np.isfinite(gap):
        ax.set_ylim(-gap, offsets[-1] + gap)
        ax.set_xlim(tt[0], tt[-1])
        ax.set_yticks(offsets)
        ax.set_yticklabels(np.arange(trace_window[0], trace_window[1]))

    if title is not None:
        ax.set_title(title)

    if ylabel is not None:
        ax.set_ylabel(ylabel)

    return fig


def show_timeseries(node, **kwargs):
    if len(node.data.shape) == 1:
        return SingleTracePlotlyWidget(node, **kwargs)
    elif len(node.data.shape) == 2:
        return BaseGroupedTraceWidget(node, **kwargs)
    else:
        raise ValueError(
            "Visualization for TimeSeries that has data with shape {} not implemented".format(
                node.data.shape
            )
        )


class AbstractTraceWidget(widgets.VBox):
    def __init__(
        self,
        timeseries: TimeSeries,
        foreign_time_window_controller: StartAndDurationController = None,
        **kwargs
    ):
        super().__init__()
        self.timeseries = timeseries
        self.foreign_time_window_controller = foreign_time_window_controller
        self.controls = {}
        self.out_fig = None

        if foreign_time_window_controller is None:
            tmin = get_timeseries_mint(timeseries)
            tmax = get_timeseries_maxt(timeseries)
            self.time_window_controller = StartAndDurationController(tmax, tmin)
        else:
            self.time_window_controller = foreign_time_window_controller

        self.set_controls(**kwargs)
        self.set_out_fig()
        self.set_children()

    @abstractmethod
    def set_out_fig(self):
        return

    def set_controls(self, **kwargs):
        self.controls.update(
            timeseries=fixed(self.timeseries), time_window=self.time_window_controller
        )
        self.controls.update({key: widgets.fixed(val) for key, val in kwargs.items()})

    def set_children(self):
        if self.foreign_time_window_controller:
            self.children = [self.out_fig]
        else:
            self.children = [self.time_window_controller, self.out_fig]


class SingleTracePlotlyWidget(AbstractTraceWidget):
    def __init__(
        self,
        timeseries: TimeSeries,
        foreign_time_window_controller: StartAndDurationController = None,
        **kwargs
    ):
        super().__init__(
            timeseries=timeseries,
            foreign_time_window_controller=foreign_time_window_controller,
            **kwargs,
        )

    def set_out_fig(self):
        timeseries = self.controls["timeseries"].value
        time_window = self.controls["time_window"].value
        self.out_fig = show_indexed_timeseries_plotly(
            timeseries=timeseries, time_window=time_window
        )

        def on_change(change):
            time_window = self.controls["time_window"].value
            istart = timeseries_time_to_ind(timeseries, time_window[0])
            istop = timeseries_time_to_ind(timeseries, time_window[1])
            yy, units = get_timeseries_in_units(timeseries, istart, istop)
            self.out_fig.data[0].x = get_timeseries_tt(timeseries, istart, istop)
            self.out_fig.data[0].y = list(yy)

            # Get data y-range, catching case with no data in current range (if so - no update)
            y_range = [min(yy), max(yy)] if yy.size != 0 else [None, None]
            self.out_fig.update_layout(
                yaxis={"range": y_range, "autorange": False},
                xaxis={"range": time_window, "autorange": False},
            )

        self.controls["time_window"].observe(on_change)


class SeparateTracesPlotlyWidget(AbstractTraceWidget):
    def set_out_fig(self):

        timeseries = self.controls["timeseries"].value

        time_window = self.controls["time_window"].value

        if len(timeseries.data.shape) > 1:
            color = DEFAULT_PLOTLY_COLORS
            no_rows = timeseries.data.shape[1]
            self.out_fig = go.FigureWidget(make_subplots(rows=no_rows, cols=1))

            for i, xyz in enumerate(("x", "y", "z")[:no_rows]):
                self.out_fig = show_indexed_timeseries_plotly(
                    timeseries=timeseries,
                    time_window=time_window,
                    trace_range=[i, i + 1],
                    fig=self.out_fig,
                    col=1,
                    row=i + 1,
                    scatter_kwargs=dict(marker_color=color[i % len(color)], name=xyz),
                    figure_kwargs=dict(yaxis=dict(title_text=xyz)),
                )
        else:
            self.out_fig = show_indexed_timeseries_plotly(
                timeseries=timeseries, time_window=time_window
            )

        def on_change(change):
            time_window = self.controls["time_window"].value
            istart = timeseries_time_to_ind(timeseries, time_window[0])
            istop = timeseries_time_to_ind(timeseries, time_window[1])

            tt = get_timeseries_tt(timeseries, istart, istop)
            yy, units = get_timeseries_in_units(timeseries, istart, istop)

            with self.out_fig.batch_update():
                if len(yy.shape) == 1:
                    self.out_fig.data[0].x = tt
                    self.out_fig.data[0].y = yy
                    self.out_fig.update_yaxes(range=[min(yy), max(yy)], row=1, col=1)
                    self.out_fig.update_xaxes(range=[min(tt), max(tt)], row=1, col=1)
                else:
                    for i, dd in enumerate(yy.T):
                        self.out_fig.data[i].x = tt
                        self.out_fig.data[i].y = dd
                        self.out_fig.update_yaxes(
                            range=[min(dd), max(dd)] if dd.size != 0 else [None, None],
                            row=i + 1, col=1
                        )
                        self.out_fig.update_xaxes(
                            range=time_window, row=i + 1, col=1
                        )

        self.controls["time_window"].observe(on_change)


def _prep_timeseries(time_series: TimeSeries, time_window=None, order=None):
    """Pull dataset region from entire dataset. Return tt and offests used for plotting

    Parameters
    ----------
    time_series: TimeSeries
    time_window
    order

    Returns
    -------

    """
    if time_window is None:
        t_ind_start = 0
        t_ind_stop = None
    else:
        t_ind_start = timeseries_time_to_ind(time_series, time_window[0])
        t_ind_stop = timeseries_time_to_ind(time_series, time_window[1])

    tt = get_timeseries_tt(time_series, t_ind_start, t_ind_stop)
    unique_sorted_order, inverse_sort = np.unique(order, return_inverse=True)

    if len(time_series.data.shape) > 1:
        mini_data = time_series.data[t_ind_start:t_ind_stop, unique_sorted_order][
            :, inverse_sort
        ]
        if np.all(np.isnan(mini_data)):
            return None, tt, None
        gap = np.median(np.nanstd(mini_data, axis=0)) * 20
        offsets = np.arange(len(order)) * gap
        mini_data = mini_data + offsets
    else:
        mini_data = time_series.data[t_ind_start:t_ind_stop]
        offsets = [0]

    return mini_data, tt, offsets


def plot_grouped_traces(
    time_series: TimeSeries,
    time_window=None,
    order=None,
    ax=None,
    figsize=(8, 7),
    group_inds=None,
    labels=None,
    colors=color_wheel,
    show_legend=True,
    dynamic_table_region_name=None,
    window=None,
    **kwargs
):
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)

    if order is None:
        if len(time_series.data.shape) > 1:
            order = np.arange(time_series.data.shape[1])
        else:
            order = [0]

    if group_inds is not None:
        row_ids = getattr(time_series, dynamic_table_region_name).data[:]
        channel_inds = [np.argmax(row_ids == x) for x in order]
    elif window is not None:
        order = order[window[0] : window[1]]
        channel_inds = order
    else:
        channel_inds = order

    if len(channel_inds):
        mini_data, tt, offsets = _prep_timeseries(time_series, time_window, channel_inds)
    else:
        mini_data = None
        tt = time_window

    if mini_data is None:
        ax.plot(tt, np.ones_like(tt) * np.nan, color="k")
        return

    if group_inds is not None:
        ugroup_inds = np.unique(group_inds)
        handles = []

        for i, ui in enumerate(ugroup_inds):
            color = colors[ugroup_inds[i] % len(colors)]
            lines_handle = ax.plot(tt, mini_data[:, group_inds == ui], color=color)
            handles.append(lines_handle[0])

        if show_legend:
            ax.legend(
                handles=handles[::-1],
                labels=list(labels[ugroup_inds][::-1]),
                loc="upper left",
                bbox_to_anchor=(1.01, 1),
            )
    else:
        ax.plot(tt, mini_data, color="k")

    ax.set_xlim((tt[0], tt[-1]))
    ax.set_xlabel("time (s)")

    if len(offsets) > 1:
        ax.set_ylim(
            offsets[0] - (offsets[1] - offsets[0]) / 2,
            offsets[-1] + (offsets[-1] - offsets[-2]) / 2,
        )
    if len(order) <= 30:
        ax.set_yticks(offsets)
        ax.set_yticklabels(order)
    else:
        ax.set_yticks([])


def plot_grouped_traces_plotly(
    time_series: TimeSeries,
    time_window,
    order,
    group_inds=None,
    labels=None,
    colors=color_wheel,
    fig=None,
    **kwargs
):
    mini_data, tt, offsets = _prep_timeseries(time_series, time_window, order)

    if fig is None:
        fig = go.FigureWidget()
    if group_inds is not None:
        ugroup_inds = np.unique(group_inds)
        for igroup, ui in enumerate(ugroup_inds[::-1]):
            color = colors[ugroup_inds[::-1][igroup] % len(colors)]
            group_data = mini_data[:, group_inds == ui].T
            multi_trace(tt, group_data, color, labels[ui], fig=fig)
    else:
        multi_trace(tt, mini_data.T, "black", fig=fig)
    fig.update_layout(title=time_series.name, xaxis_title="time (s)")
    fig.update_yaxes(tickvals=list(offsets), ticktext=[str(i) for i in order])

    return fig


class BaseGroupedTraceWidget(widgets.HBox):
    def __init__(
        self,
        time_series: TimeSeries,
        dynamic_table_region_name=None,
        foreign_time_window_controller: StartAndDurationController = None,
        foreign_group_and_sort_controller: GroupAndSortController = None,
        mpl_plotter=plot_grouped_traces,
        **kwargs
    ):
        """

        Parameters
        ----------
        time_series: TimeSeries
        dynamic_table_region_name: str, optional
        foreign_time_window_controller: StartAndDurationController, optional
        foreign_group_and_sort_controller: GroupAndSortController, optional
        mpl_plotter: function
            Choose function to use when creating figures
        kwargs
        """

        if (
            dynamic_table_region_name is not None
            and foreign_group_and_sort_controller is not None
        ):
            raise TypeError(
                "You cannot supply both `dynamic_table_region_name` and `foreign_group_and_sort_controller`."
            )

        super().__init__()
        self.time_series = time_series

        if foreign_time_window_controller is not None:
            self.time_window_controller = foreign_time_window_controller
        else:
            self.tmin = get_timeseries_mint(time_series)
            self.tmax = get_timeseries_maxt(time_series)
            self.time_window_controller = StartAndDurationController(
                tmin=self.tmin, tmax=self.tmax
            )

        self.controls = dict(
            time_series=widgets.fixed(self.time_series),
            time_window=self.time_window_controller,
            dynamic_table_region_name=widgets.fixed(dynamic_table_region_name),
        )
        if foreign_group_and_sort_controller is None:
            if dynamic_table_region_name is not None:
                dynamic_table_region = getattr(time_series, dynamic_table_region_name)
                table = dynamic_table_region.table
                referenced_rows = np.array(dynamic_table_region.data)
                self.gas = GroupAndSortController(
                    dynamic_table=table,
                    keep_rows=referenced_rows,
                )
                self.controls.update(gas=self.gas)
            else:
                self.gas = None
                range_controller_max = min(30, self.time_series.data.shape[1])
                self.range_controller = RangeController(
                    0,
                    self.time_series.data.shape[1],
                    start_value=(0, range_controller_max),
                    dtype="int",
                    description="traces",
                    orientation="vertical",
                )
                self.controls.update(window=self.range_controller)
        else:
            self.gas = foreign_group_and_sort_controller
            self.controls.update(gas=self.gas)

        # Sets up interactive output controller
        out_fig = interactive_output(mpl_plotter, self.controls)

        if foreign_time_window_controller:
            right_panel = out_fig
        else:
            right_panel = widgets.VBox(
                children=[
                    self.time_window_controller,
                    out_fig,
                ],
                layout=widgets.Layout(width="100%"),
            )

        if foreign_group_and_sort_controller or self.gas is None:
            if self.range_controller is None:
                self.children = [right_panel]
            else:
                self.children = [self.range_controller, right_panel]
        else:

            self.children = [self.gas, right_panel]

        # self.layout = widgets.Layout(width="100%")


class MultiTimeSeriesWidget(widgets.VBox):
    def __init__(self, time_series_list, widget_class_list, constrain_time_range=False):
        """

        Parameters
        ----------
        time_series_list: list of TimeSeries
        widget_class_list: list of classes, optional
        constrain_time_range: bool, optional
            Default is False
        """
        super().__init__()
        if constrain_time_range:
            self.tmin = max(
                get_timeseries_mint(time_series) for time_series in time_series_list
            )
            self.tmax = min(
                get_timeseries_maxt(time_series) for time_series in time_series_list
            )
        else:
            self.tmin = min(
                get_timeseries_mint(time_series) for time_series in time_series_list
            )
            self.tmax = max(
                get_timeseries_maxt(time_series) for time_series in time_series_list
            )
        self.time_window_controller = StartAndDurationController(
            tmin=self.tmin, tmax=self.tmax
        )

        ts_widgets = [
            widget_class(
                time_series, foreign_time_window_controller=self.time_window_controller
            )
            for widget_class, time_series in zip(widget_class_list, time_series_list)
        ]
        self.children = [self.time_window_controller] + ts_widgets


class AlignMultiTraceTimeSeriesByTrialsAbstract(widgets.VBox):
    def __init__(
        self,
        time_series: TimeSeries,
        trials: TimeIntervals = None,
        trace_index=0,
        trace_controller=None,
        trace_controller_kwargs=None,
        sem=True,
    ):

        self.time_series = time_series
        self.time_series_data = time_series.data[()]
        self.time_series_timestamps = None
        if time_series.rate is None:
            self.time_series_timestamps = time_series.timestamps[()]
        super().__init__()

        if trials is None:
            self.trials = self.get_trials()
            if self.trials is None:
                self.children = [widgets.HTML("No trials present")]
                return
        else:
            self.trials = trials

        if trace_controller is None:
            ntraces = self.time_series.data.shape[1]
            input_trace_controller_kwargs = dict(
                options=[x for x in range(ntraces)],
                value=trace_index,
                description="trace",
                layout=Layout(width="200px"),
            )
            if trace_controller_kwargs is not None:
                input_trace_controller_kwargs.update(trace_controller_kwargs)
            self.trace_controller = widgets.Dropdown(**input_trace_controller_kwargs)
        else:
            self.trace_controller = trace_controller

        self.trial_event_controller = make_trial_event_controller(
            self.trials, layout=Layout(width="200px")
        )

        self.before_ft = widgets.FloatText(
            0.5, min=0, description="before (s)", layout=Layout(width="200px")
        )
        self.after_ft = widgets.FloatText(
            2.0, min=0, description="after (s)", layout=Layout(width="200px")
        )

        self.gas = self.make_group_and_sort(window=False, control_order=False)

        self.align_to_zero_cb = widgets.Checkbox(description="align to zero")

        self.controls = dict(
            index=self.trace_controller,
            after=self.after_ft,
            before=self.before_ft,
            start_label=self.trial_event_controller,
            gas=self.gas,
            align_to_zero=self.align_to_zero_cb,
        )
        vbox_cols = [
            [self.gas, self.align_to_zero_cb],
            [
                self.trace_controller,
                self.trial_event_controller,
                self.before_ft,
                self.after_ft,
            ],
        ]
        if sem:
            self.sem_cb = widgets.Checkbox(description="show SEM")
            self.controls.update(sem=self.sem_cb)
            vbox_cols[0].append(self.sem_cb)
        out_fig = set_plotly_callbacks(self.update, self.controls)

        self.children = [widgets.HBox([widgets.VBox(i) for i in vbox_cols]), out_fig]

    def get_trials(self):
        return self.time_series.get_ancestor("NWBFile").trials

    def make_group_and_sort(
        self, window=None, control_order=False, control_limit=False
    ):
        return GroupAndSortController(
            self.trials,
            window=window,
            control_order=control_order,
            control_limit=control_limit,
        )

    def plot_group(self,group_inds, data_trialized, time_trialized, fig, order):
        for group in np.unique(group_inds):
            line_color = color_wheel[group%len(color_wheel)]
            pb = ProgressBar(np.where(group_inds==group)[0],
                             desc=f"plotting {group} data", leave=False)
            group_data = []
            group_ts = []
            for i,trim_trial_no in enumerate(pb):
                trial_idx = order[trim_trial_no]
                group_data.append(data_trialized[trial_idx])
                group_ts.append(time_trialized[trial_idx])
            fig = multi_trace(group_ts,group_data,fig=fig,color=line_color,label=str(group),insert_nans=True)
        tt_flat = np.concatenate(time_trialized)
        fig.update_layout(xaxis_title='time (s)',
                          yaxis_title=self.time_series.name,
                          xaxis_range=(np.min(tt_flat), np.max(tt_flat)))
        return fig


class AlignMultiTraceTimeSeriesByTrialsConstant(
    AlignMultiTraceTimeSeriesByTrialsAbstract
):
    def __init__(
        self,
        time_series: TimeSeries,
        trials: TimeIntervals = None,
        trace_index=0,
        trace_controller=None,
        trace_controller_kwargs=None,
    ):

        self.time_series = time_series

        super().__init__(
            time_series=time_series,
            trials=trials,
            trace_index=trace_index,
            trace_controller=trace_controller,
            trace_controller_kwargs=trace_controller_kwargs,
            sem=True,
        )

    @functools.lru_cache()
    def align_data(self, start_label, before, after, index=None):
        starts = np.array(self.trials[start_label][:]) - before
        out_data_aligned = []
        out_ts_aligned = []
        for start in starts:
            idx_start = int((start - self.time_series.starting_time)*self.time_series.rate)
            idx_stop = int(idx_start + (before+after)*self.time_series.rate)
            out_ts_aligned.append(np.linspace(-before,after,num=idx_stop-idx_start))
            if len(self.time_series_data.shape) > 1 and index is not None:
                out_data_aligned.append(self.time_series_data[idx_start:idx_stop, index])
            else:
                out_data_aligned.append(self.time_series_data[idx_start:idx_stop])
        return out_data_aligned, out_ts_aligned

    def update(
        self,
        index: int,
        start_label: str = "start_time",
        before: float = 0.0,
        after: float = 1.0,
        order=None,
        group_inds=None,
        labels=None,
        align_to_zero=False,
        sem=False,
        fig:go.FigureWidget = None,
    ):
        data, time_ts_aligned = self.align_data(start_label, before, after, index)
        if group_inds is None:
            group_inds = np.zeros(len(self.trials), dtype=int)
        if align_to_zero:
            for trial_no in order:
                data_zero_id = bisect(time_ts_aligned[trial_no], 0)
                data[trial_no] -= data[trial_no][data_zero_id]
        fig = go.FigureWidget() if fig is None else fig
        fig.data = []
        fig.layout = {}
        if sem:
            group_stats = []
            for group in np.unique(group_inds):
                this_mean = np.nanmean(data[group_inds == group, :], axis=0)
                err = scipy.stats.sem(
                    data[group_inds == group, :], axis=0, nan_policy="omit"
                )
                group_stats.append(
                    dict(
                        mean=this_mean,
                        lower=this_mean - 2 * err,
                        upper=this_mean + 2 * err,
                        group=group,
                    )
                )

            for stats in group_stats:
                plot_kwargs = dict()
                color = color_wheel[stats["group"]]
                if labels is not None:
                    plot_kwargs.update(text=labels[stats["group"]])
                fig.add_scattergl(x=time_ts_aligned[0], y=stats["lower"], line_color=color)
                fig.add_scattergl(x=time_ts_aligned[0], y=stats["upper"], line_color=color, fill='tonexty', opacity=0.2)
                fig.add_scattergl(x=time_ts_aligned[0], y=stats["mean"], line_color=color, **plot_kwargs)

        else:
            fig = self.plot_group(group_inds,data,time_ts_aligned,fig,order)
        return fig

class AlignMultiTraceTimeSeriesByTrialsVariable(
    AlignMultiTraceTimeSeriesByTrialsAbstract
):
    def __init__(
        self,
        time_series: TimeSeries,
        trials: TimeIntervals = None,
        trace_index=0,
        trace_controller=None,
        trace_controller_kwargs=None,
    ):

        self.time_series = time_series

        super().__init__(
            time_series=time_series,
            trials=trials,
            trace_index=trace_index,
            trace_controller=trace_controller,
            trace_controller_kwargs=trace_controller_kwargs,
            sem=False,
        )

    @functools.lru_cache()
    def align_data(self, start_label, before, after, index=None):
        starts = np.array(self.trials[start_label][:]) - before
        out_data_aligned = []
        out_ts_aligned = []
        for start in starts:
            idx_start = bisect(self.time_series_timestamps, start-before)
            idx_stop = bisect(self.time_series_timestamps, start + after, lo=idx_start)
            out_ts_aligned.append(self.time_series_timestamps[idx_start:idx_stop]-
                                  self.time_series_timestamps[idx_start]-before)
            if len(self.time_series_data.shape) > 1 and index is not None:
                out_data_aligned.append(self.time_series_data[idx_start:idx_stop, index])
            else:
                out_data_aligned.append(self.time_series_data[idx_start:idx_stop])
        return out_data_aligned, out_ts_aligned

    def update(
        self,
        index: int,
        start_label: str = "start_time",
        before: float = 0.0,
        after: float = 1.0,
        order=None,
        group_inds=None,
        labels=None,
        align_to_zero=False,
        fig:go.FigureWidget = None,
    ):

        data,time_ts_aligned = self.align_data(start_label,before,after,index)
        if group_inds is None:
            group_inds = np.zeros(len(self.trials), dtype=int)
        if align_to_zero:
            for trial_no in order:
                data_zero_id = bisect(time_ts_aligned[trial_no], 0)
                data[trial_no] -= data[trial_no][data_zero_id]
        fig = fig if fig is not None else go.FigureWidget()
        fig.data = []
        fig.layout = {}
        return self.plot_group(group_inds,data,time_ts_aligned,fig,order)



import unittest

import numpy as np

from pynwb import TimeSeries
from pynwb.epoch import TimeIntervals
from pynwb.behavior import Position, SpatialSeries, BehavioralEvents

from nwbwidgets.behavior import (
    show_behavioral_events,
    show_spatial_series,
    trial_align_spatial_series,
    SpatialSeriesTraceWidget2D,
    SpatialSeriesTraceWidget3D,
)
from nwbwidgets.view import default_neurodata_vis_spec
from nwbwidgets.base import show_multi_container_interface
from nwbwidgets.timeseries import SeparateTracesPlotlyWidget


class ShowSpatialSeriesTestCase(unittest.TestCase):
    def setUp(self):
        self.spatial_series = SpatialSeries(
            name="position",
            data=np.linspace(0, 1, 20),
            rate=50.0,
            reference_frame="starting gate",
        )

    def test_show_position(self):
        position = Position(spatial_series=self.spatial_series)

        show_multi_container_interface(position, default_neurodata_vis_spec)

    def test_show_spatial_series_over_time(self):
        SeparateTracesPlotlyWidget(self.spatial_series)

    def test_show_spatial_series(self):
        show_spatial_series(self.spatial_series)


class ShowSpatialSeriesTwoDTestCase(unittest.TestCase):
    def setUp(self):
        self.spatial_series = SpatialSeries(
            name="position",
            data=np.array([np.linspace(0, 1, 20), np.linspace(0, 1, 20)]).T,
            rate=50.0,
            reference_frame="starting gate",
        )

    def test_show_spatial_series_over_time_twoD(self):
        widget = SeparateTracesPlotlyWidget(self.spatial_series)
        widget.controls["time_window"].value = [2.0 / 50, 10.0 / 50]

    def test_show_spatial_series_twoD(self):
        widget = SpatialSeriesTraceWidget2D(self.spatial_series)
        widget.controls["time_window"].value = [2.0 / 50, 10.0 / 50]


class ShowSpatialSeriesThreeDTestCase(unittest.TestCase):
    def setUp(self):
        self.spatial_series = SpatialSeries(
            name="position",
            data=np.array(
                [np.linspace(0, 1, 20), np.linspace(0, 1, 20), np.linspace(0, 1, 20)]
            ).T,
            rate=50.0,
            reference_frame="starting gate",
        )

    def test_show_spatial_series_over_time_threeD(self):
        widget = SeparateTracesPlotlyWidget(self.spatial_series)
        widget.controls["time_window"].value = [2.0 / 50, 10.0 / 50]

    def test_show_spatial_series_threeD(self):
        widget = SpatialSeriesTraceWidget3D(self.spatial_series)
        widget.controls["time_window"].value = [2.0 / 50, 10.0 / 50]


class SpatialSeriesTrialsAlign(unittest.TestCase):
    def setUp(self) -> None:
        data = np.random.rand(100, 3)
        timestamps = [0.0]
        for _ in range(data.shape[0]):
            timestamps.append(timestamps[-1] + 0.75 + 0.25 * np.random.rand())
        self.spatial_series_rate = SpatialSeries(
            name="position_rate",
            data=data,
            starting_time=0.0,
            rate=1.0,
            reference_frame="starting gate",
        )
        self.spatial_series_ts = SpatialSeries(
            name="position_ts",
            data=data,
            timestamps=np.array(timestamps),
            reference_frame="starting gate",
        )
        self.time_intervals = TimeIntervals(name="Test Time Interval")
        n_intervals = 10
        for start_time in np.linspace(0, 75, n_intervals + 1):
            if start_time < 75:
                stt = start_time + np.random.rand()
                spt = stt + 7 - np.random.rand()
                self.time_intervals.add_interval(start_time=stt, stop_time=spt)
        self.time_intervals.add_column(
            name="temp", description="desc", data=np.random.randint(2, size=n_intervals)
        )
        self.time_intervals.add_column(
            name="temp2",
            description="desc",
            data=np.random.randint(10, size=n_intervals),
        )

    def test_spatial_series_trials_align_rate(self):
        trial_align_spatial_series(self.spatial_series_rate, self.time_intervals)

    def test_spatial_series_trials_align_ts(self):
        trial_align_spatial_series(self.spatial_series_ts, self.time_intervals)


def test_show_behavioral_events():
    data = np.arange(100, 200, 10)
    ts = TimeSeries(
        name="test_timeseries", data=data, unit="m", starting_time=0.0, rate=1.0
    )

    beh_events = BehavioralEvents(time_series=ts)

    show_behavioral_events(beh_events, default_neurodata_vis_spec)



from datetime import datetime
from dateutil.tz import tzlocal

import ipywidgets as widgets

from pynwb import NWBFile

from nwbwidgets import nwb2widget


def test_nwbfile():
    nwbfile = NWBFile(
        "my first synthetic recording",
        "EXAMPLE_ID",
        datetime.now(tzlocal()),
        experimenter="Dr. Matthew Douglass",
        lab="Vision Neuroscience Laboratory",
        institution="University of Middle Earth at the Shire",
        experiment_description="We recorded from two macaque monkeys during memory-guided saccade task",
        session_id="LONELYMTL",
    )

    assert isinstance(nwb2widget(nwbfile), widgets.Widget)



import unittest

from datetime import datetime
from dateutil.tz import tzlocal

import numpy as np

import ipywidgets as widgets

from pynwb import NWBFile
from pynwb.epoch import TimeIntervals

from nwbwidgets.utils.units import (
    get_min_spike_time,
    align_by_trials,
    align_by_time_intervals,
)
from nwbwidgets.base import TimeIntervalsSelector
from nwbwidgets.misc import TuningCurveWidget, TuningCurveExtendedWidget


class UnitsTrialsTestCase(unittest.TestCase):

    def setUp(self):
        start_time = datetime(2017, 4, 3, 11, tzinfo=tzlocal())
        create_date = datetime(2017, 4, 15, 12, tzinfo=tzlocal())

        self.nwbfile = NWBFile(
            session_description="NWBFile for PSTH",
            identifier="NWB123",
            session_start_time=start_time,
            file_create_date=create_date,
        )

        self.nwbfile.add_unit_column("location", "the anatomical location of this unit")
        self.nwbfile.add_unit_column(
            "quality", "the quality for the inference of this unit"
        )

        self.nwbfile.add_unit(
            id=1,
            spike_times=[2.2, 3.0, 4.5],
            obs_intervals=[[1, 10]],
            location="CA1",
            quality=0.95,
        )
        self.nwbfile.add_unit(
            id=2,
            spike_times=[2.2, 3.0, 25.0, 26.0],
            obs_intervals=[[1, 10], [20, 30]],
            location="CA3",
            quality=0.85,
        )
        self.nwbfile.add_unit(
            id=3,
            spike_times=[1.2, 2.3, 3.3, 4.5],
            obs_intervals=[[1, 10], [20, 30]],
            location="CA1",
            quality=0.90,
        )

        self.nwbfile.add_trial_column(
            name="stim", description="the visual stimuli during the trial"
        )

        self.nwbfile.add_trial(start_time=0.0, stop_time=2.0, stim="person")
        self.nwbfile.add_trial(start_time=3.0, stop_time=5.0, stim="ocean")
        self.nwbfile.add_trial(start_time=6.0, stop_time=8.0, stim="desert")
        self.nwbfile.add_trial(start_time=8.0, stop_time=12.0, stim="person")
        self.nwbfile.add_trial(start_time=13.0, stop_time=15.0, stim="ocean")
        self.nwbfile.add_trial(start_time=16.0, stop_time=18.0, stim="desert")



class ExtendedTimeIntervalSelector(TimeIntervalsSelector):
    InnerWidget = TuningCurveWidget


class ExtendedTimeIntervalSelectorTestCase(UnitsTrialsTestCase):

    def setUp(self):
        super().setUp()

        # add intervals to nwbfile
        ti1 = TimeIntervals(name='intervals', description='experimental intervals')
        ti1.add_interval(start_time=0.0, stop_time=2.0)
        ti1.add_interval(start_time=2.0, stop_time=4.0)
        ti1.add_interval(start_time=4.0, stop_time=6.0)
        ti1.add_interval(start_time=6.0, stop_time=8.0)
        ti1.add_column(name='var1', data=['a', 'b', 'a', 'b'], description='no description')
        self.nwbfile.add_time_intervals(ti1)

        self.widget = ExtendedTimeIntervalSelector(
            input_data=self.nwbfile.units
        )

    def test_make_widget(self):
        assert isinstance(self.widget, widgets.Widget)

    def test_widget_children(self):
        assert len(self.widget.children) == 2

        for i, c in enumerate(self.widget.children):
            assert isinstance(c, widgets.Widget), f'{i}th child of TuningCurve widget is not a widget'

    def test_make_graph(self):
        # rows controller triggers drawing of graphic
        self.widget.children[1].children[1].value = 'var1'

        for i, c in enumerate(self.widget.children):
            assert isinstance(c, widgets.Widget), f'{i}th child of TuningCurve widget is not a widget'


class TuningCurveTestCase(UnitsTrialsTestCase):

    def setUp(self):
        super().setUp()
        self.widget = TuningCurveWidget(
            units=self.nwbfile.units,
            trials=self.nwbfile.trials
        )
        # rows controller triggers drawing of graphic
        self.widget.children[0].children[1].value = 'stim'

    def test_make_widget(self):
        assert isinstance(self.widget, widgets.Widget)

    def test_widget_children(self):
        assert len(self.widget.children) == 2

        for i, c in enumerate(self.widget.children):
            assert isinstance(c, widgets.Widget), f'{i}th child of TuningCurve widget is not a widget'


class TuningCurveRasterGridCombinedTestCase(UnitsTrialsTestCase):

    def setUp(self):
        super().setUp()
        self.widget = TuningCurveExtendedWidget(
            units=self.nwbfile.units,
            trials=self.nwbfile.trials
        )

    def test_make_widget(self):
        assert isinstance(self.widget, widgets.Widget)

    def test_widget_children(self):
        assert len(self.widget.children) == 3

        for i, c in enumerate(self.widget.children):
            assert isinstance(c, widgets.Widget), f'{i}th child of TuningCurve widget is not a widget'



class ShowPSTHTestCase(UnitsTrialsTestCase):

    def test_get_min_spike_time(self):
        assert get_min_spike_time(self.nwbfile.units) == 1.2

    def test_align_by_trials(self):
        compare_to_at = [
            np.array([2.2, 3.0, 25.0, 26.0]),
            np.array([-0.8, 0.0, 22.0, 23.0]),
            np.array([-3.8, -3.0, 19.0, 20.0]),
            np.array([-5.8, -5., 17., 18.]),
            np.array([-10.8, -10.,  12.,  13.]),
            np.array([-13.8, -13.,   9.,  10.])
        ]

        at = align_by_trials(
            self.nwbfile.units,
            index=1,
            start=-20.0,
            end=30.0
        )

        np.testing.assert_allclose(at, compare_to_at, rtol=1e-02)

    def test_align_by_time_intervals_Nonetrials_select(self):
        time_intervals = TimeIntervals(name="Test Time Interval")
        time_intervals.add_interval(start_time=21.0, stop_time=28.0)
        time_intervals.add_interval(start_time=22.0, stop_time=26.0)
        time_intervals.add_interval(start_time=22.0, stop_time=28.4)

        ati = align_by_time_intervals(
            self.nwbfile.units,
            index=1,
            intervals=time_intervals,
            stop_label=None,
            start=-20.0,
            end=30.0,
        )

        compare_to_ati = [
            np.array([-18.8, -18.0, 4.0, 5.0]),
            np.array([-19.8, -19.0, 3.0, 4.0]),
            np.array([-19.8, -19.0, 3.0, 4.0]),
        ]

        np.testing.assert_array_equal(ati, compare_to_ati)

    def test_align_by_time_intervals(self):
        time_intervals = TimeIntervals(name="Test Time Interval")
        time_intervals.add_interval(start_time=21.0, stop_time=28.0)
        time_intervals.add_interval(start_time=22.0, stop_time=26.0)
        time_intervals.add_interval(start_time=22.0, stop_time=28.4)

        ati = align_by_time_intervals(
            self.nwbfile.units,
            index=1,
            intervals=time_intervals,
            stop_label=None,
            start=-20.0,
            end=30.0,
            rows_select=[0, 1],
        )

        compare_to_ati = [
            np.array([-18.8, -18.0, 4.0, 5.0]),
            np.array([-19.8, -19.0, 3.0, 4.0]),
        ]

        np.testing.assert_array_equal(ati, compare_to_ati)



from bisect import bisect, bisect_left

import numpy as np

from pynwb import TimeSeries


def get_timeseries_tt(node: TimeSeries, istart=0, istop=None) -> np.ndarray:
    """
    For any TimeSeries, return timestamps. If the TimeSeries uses starting_time and rate, the timestamps will be
    generated.

    Parameters
    ----------
    node: pynwb.TimeSeries
    istart: int, optional
        Optionally sub-select the returned times - lower bound
    istop: int, optional
        Optionally sub-select the returned times - upper bound

    Returns
    -------
    numpy.ndarray

    """
    if node.timestamps is not None:
        return node.timestamps[istart:istop]
    else:
        if not np.isfinite(node.starting_time):
            starting_time = 0
        else:
            starting_time = node.starting_time
        if istop is None:
            return np.arange(istart, len(node.data)) / node.rate + starting_time
        elif istop > 0:
            return np.arange(istart, istop) / node.rate + starting_time
        else:
            return (
                np.arange(istart, len(node.data) + istop - 1) / node.rate
                + starting_time
            )


def get_timeseries_maxt(node: TimeSeries) -> float:
    """
    Returns the maximum time of any TimeSeries

    Parameters
    ----------
    node: pynwb.TimeSeries

    Returns
    -------
    float

    """
    if node.timestamps is not None:
        return node.timestamps[-1]
    elif np.isnan(node.starting_time):
        return (len(node.data) - 1) / node.rate
    else:
        return (len(node.data) - 1) / node.rate + node.starting_time


def get_timeseries_mint(node: TimeSeries) -> float:
    """
    Returns the minimum time of any TimeSeries

    Parameters
    ----------
    node: pynwb.TimeSeries

    Returns
    -------
    float

    """
    if node.timestamps is not None:
        return node.timestamps[0]
    elif np.isnan(node.starting_time):
        return 0
    else:
        return node.starting_time


def get_timeseries_in_units(node: TimeSeries, istart=None, istop=None):
    """
    Convert data into the designated units

    Parameters
    ----------
    node: pynwb.TimeSeries
    istart: int
    istop: int

    Returns
    -------
    numpy.ndarray, str

    """
    data = node.data[istart:istop]
    if node.conversion and np.isfinite(node.conversion):
        data = data * node.conversion
        unit = node.unit
    else:
        unit = None
    return data, unit


def timeseries_time_to_ind(node: TimeSeries, time, ind_min=None, ind_max=None) -> int:
    """
    Get the index of a certain time for any TimeSeries. For TimeSeries that use timestamps, bisect is used. You can
    optionally provide ind_min and ind_max to constrain the search.

    Parameters
    ----------
    node: pynwb.TimeSeries
    time: float
    ind_min: int, optional
    ind_max: int, optional

    Returns
    -------

    """
    if node.timestamps is not None:
        kwargs = dict()
        if ind_min is not None:
            kwargs.update(lo=ind_min)
        if ind_max is not None:
            kwargs.update(hi=ind_max)
        id_found = bisect_left(node.timestamps, time, **kwargs)
        return id_found if id_found < len(node.data) else len(node.data) - 1
    else:
        if np.isnan(node.starting_time):
            starting_time = 0
        else:
            starting_time = node.starting_time
        id_found = int(np.ceil((time - starting_time) * node.rate))
        return id_found if id_found < len(node.data) else len(node.data) - 1


def bisect_timeseries_by_times(
    timeseries: TimeSeries, starts, duration: float, traces=None
):
    """
    Parameters
    ----------
    timeseries: TimeSeries
    starts: iterable
        time at which to bisect
    duration: float
        duration of window after start
    traces: int
        index into the second dim of data
    Returns
    -------
    out: list
        list with bisected arrays from data
    """
    out = []
    for start in starts:
        if timeseries.rate is not None:
            idx_start = int((start - timeseries.starting_time) * timeseries.rate)
            idx_stop = int(idx_start + duration * timeseries.rate)
        else:
            idx_start = bisect(timeseries.timestamps, start)
            idx_stop = bisect(timeseries.timestamps, start + duration, lo=idx_start)
        if len(timeseries.data.shape) > 1 and traces is not None:
            out.append(timeseries.data[idx_start:idx_stop, traces])
        else:
            out.append(timeseries.data[idx_start:idx_stop])
    return out


def align_by_times_with_timestamps(
    timeseries: TimeSeries, starts, duration: float, traces=None
):
    """
    Parameters
    ----------
    timeseries: TimeSeries
        timeseries with variable timestamps
    starts: array-like
        starts in seconds
    duration: float
        duration in seconds
    Returns
    -------
    out: list
        list: length=(n_trials); list[0]: array, shape=(n_time, ...)
    """
    assert timeseries.timestamps is not None, "supply timeseries with timestamps"
    return bisect_timeseries_by_times(timeseries, starts, duration, traces)


def align_by_times_with_rate(
    timeseries: TimeSeries, starts, duration: float, traces=None
):
    """
    Parameters
    ----------
    timeseries: TimeSeries
        timeseries with variable timestamps
    starts: array-like
        starts in seconds
    duration: float
        duration in seconds
    Returns
    -------
    out: list
        list: length=(n_trials); list[0]: array, shape=(n_time, ...)
    """
    assert timeseries.rate is not None, "supply timeseries with start_time and rate"
    return np.array(bisect_timeseries_by_times(timeseries, starts, duration, traces))


def align_timestamps_by_trials(
    timeseries: TimeSeries, starts, before: float, after: float
):
    """
    Parameters
    ----------
    timeseries: TimeSeries
        timeseries with variable timestamps
    starts: array-like
        starts in seconds
    duration: float
        duration in seconds
    Returns
    -------
    out: list
        list: length=(n_trials); list[0]: array, shape=(n_time, ...)
    """
    assert timeseries.timestamps is not None, "supply timeseries with timestamps"
    out = []
    for start in starts:
        idx_start = bisect(timeseries.timestamps, start)
        idx_stop = bisect(timeseries.timestamps, start + before + after, lo=idx_start)
        out.append(timeseries.timestamps[idx_start:idx_stop])
    return [list(np.array(i) - i[0] - before) for i in out]


def align_by_trials(
    timeseries: TimeSeries,
    start_label="start_time",
    before=0.0,
    after=1.0,
):
    """
    Args:
        timeseries: TimeSeries
        start_label: str
            default: 'start_time'
        before: float
            time after start_label in secs (positive goes back in time)
        after: float
            time after stop_label in secs (positive goes forward in time)
    Returns:
        np.array(shape=(n_trials, n_time, ...))
    """
    trials = timeseries.get_ancestor("NWBFile").trials
    return align_by_time_intervals(timeseries, trials, start_label, before, after)


def align_by_time_intervals(
    timeseries: TimeSeries,
    intervals,
    start_label="start_time",
    before=0.0,
    after=0.0,
    traces=None,
):
    """
    Args:
        timeseries: pynwb.TimeSeries
        intervals: pynwb.epoch.TimeIntervals
        start_label: str
            default: 'start_time'
        before: float
            time after start_label in secs (positive goes back in time)
        after: float
            time after stop_label in secs (positive goes forward in time)
        timestamps: bool
            if alignment uses timestamps or constant rate and starting time in TimeSeries
    Returns:
        np.array(shape=(n_trials, n_time, ...))
    """

    starts = np.array(intervals[start_label][:]) - before
    if timeseries.rate is not None:
        return align_by_times_with_rate(
            timeseries, starts, duration=after + before, traces=traces
        )
    else:
        return align_by_times_with_timestamps(
            timeseries, starts, duration=after + before, traces=traces
        )



from functools import lru_cache

import numpy as np
from skimage import measure

import ipywidgets as widgets
import plotly.graph_objects as go
import plotly.express as px

from pynwb.base import NWBDataInterface
from pynwb.ophys import (
    RoiResponseSeries,
    DfOverF,
    PlaneSegmentation,
    TwoPhotonSeries,
    ImageSegmentation,
)

from tifffile import imread, TiffFile
from ndx_grayscalevolume import GrayscaleVolume

from .base import df_to_hover_text
from .timeseries import BaseGroupedTraceWidget
from .utils.cmaps import linear_transfer_function
from .utils.dynamictable import infer_categorical_columns
from .controllers import ProgressBar

color_wheel = ["red", "blue", "green", "black", "magenta", "yellow"]


class TwoPhotonSeriesWidget(widgets.VBox):
    """Widget showing Image stack recorded over time from 2-photon microscope."""

    def __init__(self, indexed_timeseries: TwoPhotonSeries, neurodata_vis_spec: dict):
        super().__init__()

        def _add_fig_trace(img_fig: go.Figure, index):
            if self.figure is None:
                self.figure = go.FigureWidget(img_fig)
            else:
                self.figure.for_each_trace(lambda trace: trace.update(img_fig.data[0]))
            self.figure.layout.title = f"Frame no: {index}"

        if indexed_timeseries.data is None:
            if indexed_timeseries.external_file is not None:
                path_ext_file = indexed_timeseries.external_file[0]
                # Get Frames dimensions
                tif = TiffFile(path_ext_file)
                n_samples = len(tif.pages)
                page = tif.pages[0]

                def update_figure(index=0):
                    # Read first frame
                    img_fig = px.imshow(
                        imread(path_ext_file, key=int(index)), binary_string=True
                    )
                    _add_fig_trace(img_fig, index)

                slider = widgets.IntSlider(
                    value=0, min=0, max=n_samples - 1, orientation="horizontal"
                )
        else:
            if len(indexed_timeseries.data.shape) == 3:

                def update_figure(index=0):
                    img_fig = px.imshow(
                        indexed_timeseries.data[index].T, binary_string=True
                    )
                    _add_fig_trace(img_fig, index)

            elif len(indexed_timeseries.data.shape) == 4:
                import ipyvolume.pylab as p3

                output = widgets.Output()

                def update_figure(index=0):
                    p3.figure()
                    p3.volshow(
                        indexed_timeseries.data[index].transpose([1, 0, 2]),
                        tf=linear_transfer_function([0, 0, 0], max_opacity=0.3),
                    )
                    output.clear_output(wait=True)
                    self.figure = output
                    with output:
                        p3.show()

            else:
                raise NotImplementedError

            slider = widgets.IntSlider(
                value=0,
                min=0,
                max=indexed_timeseries.data.shape[0] - 1,
                orientation="horizontal",
            )

        slider.observe(lambda change: update_figure(change.new), names="value")
        self.figure = None
        self.controls = dict(slider=slider)
        update_figure()
        self.children = [self.figure, slider]


def show_df_over_f(df_over_f: DfOverF, neurodata_vis_spec: dict):
    if len(df_over_f.roi_response_series) == 1:
        title, data_input = list(df_over_f.roi_response_series.items())[0]
        return neurodata_vis_spec[RoiResponseSeries](
            data_input, neurodata_vis_spec, title=title
        )
    else:
        return neurodata_vis_spec[NWBDataInterface](df_over_f, neurodata_vis_spec)


def show_image_segmentation(img_seg: ImageSegmentation, neurodata_vis_spec: dict):
    if len(img_seg.plane_segmentations) == 1:
        return route_plane_segmentation(
            list(img_seg.plane_segmentations.values())[0], neurodata_vis_spec
        )
    else:
        return neurodata_vis_spec[NWBDataInterface](img_seg, neurodata_vis_spec)


def show_plane_segmentation_3d_voxel(plane_seg: PlaneSegmentation):
    import ipyvolume.pylab as p3

    nrois = len(plane_seg)

    voxel_mask = plane_seg["voxel_mask"]

    mx, my, mz = 0, 0, 0
    for voxel in voxel_mask:
        for x, y, z, _ in voxel:
            mx = max(mx, x)
            my = max(my, y)
            mz = max(mz, z)

    fig = p3.figure()
    for icolor, color in enumerate(color_wheel):
        vol = np.zeros((mx + 1, my + 1, mz + 1))
        sel = np.arange(icolor, nrois, len(color_wheel))
        for isel in sel:
            dat = voxel_mask[isel]
            for x, y, z, value in dat:
                vol[x, y, z] = value
        p3.volshow(vol, tf=linear_transfer_function(color, max_opacity=0.3))
    return fig


def show_plane_segmentation_3d_mask(plane_seg: PlaneSegmentation):
    import ipyvolume.pylab as p3

    nrois = len(plane_seg)

    image_masks = plane_seg["image_mask"]

    fig = p3.figure()
    for icolor, color in enumerate(color_wheel):
        vol = np.zeros(image_masks.shape[1:])
        sel = np.arange(icolor, nrois, len(color_wheel))
        for isel in sel:
            vol += plane_seg["image_mask"][isel]
        p3.volshow(vol, tf=linear_transfer_function(color, max_opacity=0.3))
    return fig


class PlaneSegmentation2DWidget(widgets.VBox):
    def __init__(self, plane_seg: PlaneSegmentation, color_wheel=color_wheel, **kwargs):
        super().__init__()
        self.categorical_columns = infer_categorical_columns(plane_seg)
        self.plane_seg = plane_seg
        self.color_wheel = color_wheel
        self.progress_bar = ProgressBar()
        self.button = widgets.Button(description="Display ROIs")
        self.children = [widgets.HBox([self.button, self.progress_bar.container])]
        self.button.on_click(self.on_button_click)
        self.kwargs = kwargs

    def on_button_click(self, b):
        if len(self.categorical_columns) == 1:
            self.color_by = list(self.categorical_columns.keys())[
                0
            ]  # changing local variables to instance variables?
            self.children += (
                self.show_plane_segmentation_2d(color_by=self.color_by, **self.kwargs),
            )
        elif len(self.categorical_columns) > 1:
            self.cat_controller = widgets.Dropdown(
                options=list(self.categorical_columns), description="color by"
            )
            self.fig = self.show_plane_segmentation_2d(
                color_by=self.cat_controller.value, **self.kwargs
            )

            def on_change(change):
                if change["new"] and isinstance(change["new"], dict):
                    ind = change["new"]["index"]
                    if isinstance(ind, int):
                        color_by = change["owner"].options[ind]
                        self.update_fig(color_by)

            self.cat_controller.observe(on_change)
            self.children += (self.cat_controller, self.fig)
        else:
            self.children += (
                self.show_plane_segmentation_2d(color_by=None, **self.kwargs),
            )
        self.children = self.children[1:]

    def update_fig(self, color_by):
        cats = np.unique(self.plane_seg[color_by][:])
        legendgroups = []
        with self.fig.batch_update():
            for color_val, data in zip(self.plane_seg[color_by][:], self.fig.data):
                color = self.color_wheel[
                    np.where(cats == color_val)[0][0]
                ]  # store the color
                data.line.color = color  # set the color
                data.legendgroup = str(color_val)  # set the legend group to the color
                data.name = str(color_val)
            for color_val, data in zip(self.plane_seg[color_by][:], self.fig.data):
                if color_val not in legendgroups:
                    data.showlegend = True
                    legendgroups.append(color_val)
                else:
                    data.showlegend = False

    def show_plane_segmentation_2d(
        self,
        color_wheel: list = color_wheel,
        color_by: str = None,
        threshold: float = 0.01,
        fig: go.Figure = None,
        width: int = 600,
        ref_image=None,
    ):
        """

        Parameters
        ----------
        plane_seg: PlaneSegmentation
        color_wheel: list, optional
        color_by: str, optional
        threshold: float, optional
        fig: plotly.graph_objects.Figure, optional
        width: int, optional
            width of image in pixels. Height is automatically determined
            to be proportional
        ref_image: image, optional


        Returns
        -------

        """
        layout_kwargs = dict()
        if color_by:
            if color_by not in self.plane_seg:
                raise ValueError(
                    "specified color_by parameter, {}, not in plane_seg object".format(
                        color_by
                    )
                )
            cats = np.unique(self.plane_seg[color_by][:])
            layout_kwargs.update(title=color_by)

        data = self.plane_seg["image_mask"].data
        n_units = len(data)
        if fig is None:
            fig = go.FigureWidget()

        if ref_image is not None:
            fig.add_trace(
                go.Heatmap(
                    z=ref_image, hoverinfo="skip", showscale=False, colorscale="gray"
                )
            )

        aux_leg = []
        import pandas as pd

        plane_seg_hover_dict = {
            key: self.plane_seg[key].data
            for key in self.plane_seg.colnames
            if key not in ["pixel_mask", "image_mask"]
        }
        plane_seg_hover_dict.update(id=self.plane_seg.id.data)
        plane_seg_hover_df = pd.DataFrame(plane_seg_hover_dict)
        all_hover = df_to_hover_text(plane_seg_hover_df)
        self.progress_bar.reset(total=n_units)
        self.progress_bar.set_description("Loading Image Masks")
        for i in range(n_units):
            kwargs = dict(showlegend=False)
            if color_by is not None:
                if plane_seg_hover_df[color_by][i] not in aux_leg:
                    kwargs.update(showlegend=True)
                    aux_leg.append(plane_seg_hover_df[color_by][i])
                c = color_wheel[np.where(cats == plane_seg_hover_df[color_by][i])[0][0]]
                kwargs.update(
                    line_color=c,
                    name=str(plane_seg_hover_df[color_by][i]),
                    legendgroup=str(plane_seg_hover_df[color_by][i]),
                )

            x, y = self.compute_outline(i, threshold)
            fig.add_trace(
                go.Scatter(
                    x=x,
                    y=y,
                    fill="toself",
                    mode="lines",
                    text=all_hover[i],
                    hovertext="text",
                    line=dict(width=0.5),
                    **kwargs,
                )
            )
            self.progress_bar.update()
        # self.progress_bar.close()
        fig.update_layout(
            width=width,
            yaxis=dict(
                mirror=True,
                scaleanchor="x",
                scaleratio=1,
                range=[0, self.plane_seg["image_mask"].shape[2]],
                constrain="domain",
            ),
            xaxis=dict(
                mirror=True,
                range=[0, self.plane_seg["image_mask"].shape[1]],
                constrain="domain",
            ),
            margin=dict(t=30, b=10),
            **layout_kwargs,
        )
        return fig

    @lru_cache(1000)
    def compute_outline(self, i, threshold):
        x, y = zip(
            *measure.find_contours(self.plane_seg["image_mask"][i], threshold)[0]
        )
        return x, y


def route_plane_segmentation(plane_seg: PlaneSegmentation, neurodata_vis_spec: dict):
    if "voxel_mask" in plane_seg:
        return show_plane_segmentation_3d_voxel(plane_seg)
    elif "image_mask" in plane_seg and len(plane_seg.image_mask.shape) == 4:
        raise NotImplementedError("3d image mask vis not implemented yet")
    elif "image_mask" in plane_seg:
        return PlaneSegmentation2DWidget(plane_seg)


def show_grayscale_volume(vol: GrayscaleVolume, neurodata_vis_spec: dict):
    import ipyvolume.pylab as p3

    fig = p3.figure()
    p3.volshow(vol.data, tf=linear_transfer_function([0, 0, 0], max_opacity=0.1))
    return fig


class RoiResponseSeriesWidget(BaseGroupedTraceWidget):
    def __init__(
        self, roi_response_series: RoiResponseSeries, neurodata_vis_spec=None, **kwargs
    ):
        super().__init__(roi_response_series, "rois", **kwargs)



from ipywidgets import widgets, Layout
import plotly.express as px

from pynwb.misc import DynamicTable

from .utils.dynamictable import infer_columns_to_plot
from .utils.widgets import interactive_output


field_lay = widgets.Layout(
        max_height="40px", max_width="500px", min_height="30px", min_width="180px"
    )


class DynamicTableSummaryWidget(widgets.VBox):
    def __init__(self, table: DynamicTable):
        super().__init__()
        self.dynamic_table = table

        self.col_names, self.categorical_cols = infer_columns_to_plot(self.dynamic_table)

        num_entries = len(self.dynamic_table)
        num_columns = len(self.dynamic_table.colnames)
        num_columns_to_plot = len(self.col_names)
        num_categorical = len(self.categorical_cols)

        self.name_text = widgets.Label(f"Table name: {self.dynamic_table.name}\n", layout=field_lay)
        self.entries_text = widgets.Label(f"Number of entries: {num_entries}\n", layout=field_lay)
        self.col_text = widgets.Label(f"Number of columns: {num_columns} - real (r): {num_columns - num_categorical}, "
                                      f"categorical (c): {num_categorical}",
                                      layout=field_lay)
        self.col_plot_text = widgets.Label(f"Number of inspectable columns: {num_columns_to_plot}")

        self.summary_text = widgets.VBox(
            [self.name_text, self.entries_text, self.col_text, self.col_plot_text])

        self.col_names_display = {}
        for col in self.col_names:
            if col in self.categorical_cols:
                self.col_names_display[f"(c) {col}"] = col
            else:
                self.col_names_display[f"(r) {col}"] = col

        self.column_dropdown = widgets.SelectMultiple(
                options=list(self.col_names_display),
                description="Inspect columns",
                layout=Layout(max_width="400px"),
                style={"description_width": "initial"},
                disabled=False,
                tooltip="Select columns to inspect. You can select at most 1 categorical and 3 real columns."
            )
        self.column_dropdown.observe(self.max_selection)

        self.nbins = widgets.IntText(
            10, min=0, description="# bins", layout=Layout(max_width="400px")
        )
        self.nbins.layout.visibility = "hidden"

        self.show_labels = widgets.Checkbox(value=True, description="show labels")

        self.plot_controls = widgets.HBox(
            [self.column_dropdown, self.nbins, self.show_labels])

        self.controls = dict(
            col_names_display=self.column_dropdown, nbins=self.nbins, show_labels=self.show_labels)

        out_fig = interactive_output(self.plot_hist_bar, self.controls)
        bottom_panel = widgets.VBox([self.plot_controls, out_fig])

        self.children = [self.summary_text, bottom_panel]

    def max_selection(self, change):
        if change['type'] == 'change' and change['name'] == 'value':
            if len(self.column_dropdown.value) > 4:
                print("Maximum number of selected items reached! "
                      "You can select at most 4 items (1 categorical and 3 real)")
                self.column_dropdown.value = ()

    def reset_dropdown(self):
        self.column_dropdown.value = ()

    def plot_hist_bar(self, col_names_display, nbins, show_labels):
        fig = None
        df = self.dynamic_table.to_dataframe()
        # remove 's' from dynamic table name
        if show_labels:
            entry_names = [f"{self.dynamic_table.name[:-1]} {index}" for index in df.index.values]
        else:
            entry_names = None

        if len(col_names_display) > 0:
            if len(col_names_display) == 1:
                # display either 1d histogram (r) or barplot (c)
                col_name_display = col_names_display[0]
                col_name = self.col_names_display[col_name_display]
                if col_name in self.categorical_cols:
                    self.nbins.layout.visibility = "hidden"
                    fig = px.histogram(df, x=col_name, histfunc='count')
                else:
                    self.nbins.layout.visibility = "visible"
                    fig = px.histogram(
                        df, x=col_name, marginal="violin", nbins=nbins)
                fig.show()
            elif len(col_names_display) == 2:
                # display either scatterplot (2x r) or multiple histograms (1r + 1c)
                real_cols = [r for r in col_names_display if "(r)" in r]
                num_real = len(real_cols)
                if num_real == 2:
                    self.nbins.layout.visibility = "hidden"
                    col_name_0 = col_name = self.col_names_display[col_names_display[0]]
                    col_name_1 = col_name = self.col_names_display[col_names_display[1]]
                    fig = px.scatter(df, x=col_name_0, y=col_name_1, marginal_x="violin", marginal_y="violin",
                                     text=entry_names)
                    fig.show()
                elif num_real == 1:
                    self.nbins.layout.visibility = "visible"
                    col_real = self.col_names_display[real_cols[0]]
                    col_cat_name = [c for c in col_names_display if c not in real_cols][0]
                    col_cat = self.col_names_display[col_cat_name]
                    fig = px.histogram(
                        df, x=col_real, color=col_cat, marginal="violin", nbins=nbins)
                    fig.show()
                else:
                    print("Select at least one real variable")
                    self.reset_dropdown()
            elif len(col_names_display) == 3:
                # display either 3d scatterplot (3x r) or colored 2d scatterplot (2r + 1c)
                real_cols = [r for r in col_names_display if "(r)" in r]
                num_real = len(real_cols)
                if num_real == 3:
                    self.nbins.layout.visibility = "hidden"
                    col_name_0 = col_name = self.col_names_display[col_names_display[0]]
                    col_name_1 = col_name = self.col_names_display[col_names_display[1]]
                    col_name_2 = col_name = self.col_names_display[col_names_display[2]]
                    fig = px.scatter_3d(
                        df, x=col_name_0, y=col_name_1, z=col_name_2,
                        text=entry_names)
                    fig.show()
                elif num_real == 2:
                    self.nbins.layout.visibility = "hidden"
                    col_real_0 = col_name = self.col_names_display[real_cols[0]]
                    col_real_1 = col_name = self.col_names_display[real_cols[1]]
                    col_cat_name = [c for c in col_names_display if c not in real_cols][0]
                    col_cat = self.col_names_display[col_cat_name]
                    df[col_cat] = df[col_cat].astype("str")
                    fig = px.scatter(df, x=col_real_0,
                                     y=col_real_1, color=col_cat,
                                     text=entry_names)
                    fig.show()
                else:
                    print("Select at most one categorical variable")
                    self.reset_dropdown()
            elif len(col_names_display) == 4:
                real_cols = [r for r in col_names_display if "(r)" in r]
                num_real = len(real_cols)

                if num_real == 3:
                    self.nbins.layout.visibility = "hidden"
                    col_name_0 = col_name = self.col_names_display[real_cols[0]]
                    col_name_1 = col_name = self.col_names_display[real_cols[1]]
                    col_name_2 = col_name = self.col_names_display[real_cols[2]]
                    col_cat_name = [c for c in col_names_display if c not in real_cols][0]
                    col_cat = self.col_names_display[col_cat_name]
                    df[col_cat] = df[col_cat].astype("str")
                    fig = px.scatter_3d(df, x=col_name_0,
                                        y=col_name_1, z=col_name_2, color=col_cat,
                                        text=entry_names)
                    fig.show()
                else:
                    print("Select 3 real and one categorical variables")
                    self.reset_dropdown()
        return fig



import numpy as np

from hdmf.common.table import VectorData
from pynwb.core import DynamicTable
from pynwb.ecephys import ElectrodeGroup
from pynwb.device import Device

from nwbwidgets.utils.dynamictable import infer_categorical_columns
from nwbwidgets.utils.testing import dicts_exact_equal


def test_infer_categorical_columns():
    data1 = np.array([1, 2, 2, 3, 1, 1, 3, 2, 3])
    data2 = np.array([3, 4, 2, 4, 3, 2, 2, 4, 4])
    device = Device(name="device")
    eg_1 = ElectrodeGroup(
        name="electrodegroup1", description="desc", location="brain", device=device
    )
    eg_2 = ElectrodeGroup(
        name="electrodegroup2", description="desc", location="brain", device=device
    )
    data3 = [eg_1, eg_2, eg_1, eg_1, eg_1, eg_1, eg_1, eg_1, eg_1]
    vd1 = VectorData("Data1", "vector data for creating a DynamicTable", data=data1)
    vd2 = VectorData("Data2", "vector data for creating a DynamicTable", data=data2)
    vd3 = VectorData(
        "ElectrodeGroup", "vector data for creating a DynamicTable", data=data3
    )
    vd = [vd1, vd2, vd3]

    dynamic_table = DynamicTable(
        name="test table",
        description="This is a test table",
        columns=vd,
        colnames=["Data1", "Data2", "ElectrodeGroup"],
    )
    assert dicts_exact_equal(
        infer_categorical_columns(dynamic_table),
        {"Data1": data1, "Data2": data2, "ElectrodeGroup": [i.name for i in data3]},
    )
